import datetime
import json
from datetime import date, timedelta
from typing import Any, AsyncGenerator, Dict, List, Tuple, Type
from unittest import mock

import pytest
import pytest_asyncio
from aiohttp import ClientSession
from aioresponses import aioresponses

from pymultimatic.api import (
    ApiError,
    defaults,
    Connector,
    WrongResponseError,
    urls,
    urls_senso,
)
from pymultimatic.model import OperatingModes, QuickModes, QuickVeto, constants, mapper
from pymultimatic.systemmanager import SystemManager, retry_async
from tests.conftest import mock_auth, path

SERIAL = mapper.map_serial_number(json.loads(open(path("files/responses/facilities")).read()))


@pytest_asyncio.fixture(name="resp", autouse=True)
async def fixture_resp(resp: aioresponses) -> AsyncGenerator[aioresponses, None]:
    with open(path("files/responses/facilities"), "r") as file:
        facilities = json.loads(file.read())
        resp.get(urls.facilities_list(), payload=facilities, status=200)
    yield resp


@pytest_asyncio.fixture(name="manager")
async def fixture_manager(
    session: ClientSession, connector: Connector
) -> AsyncGenerator[SystemManager, None]:
    manager = SystemManager("user", "pass", session, "pymultiMATIC", SERIAL)
    await connector.login()
    with mock.patch.object(connector, "request", wraps=connector.request):
        manager._connector = connector
        yield manager


@pytest_asyncio.fixture(name="senso_manager")
async def fixture_senso_manager(
    session: ClientSession, senso_connector: Connector
) -> AsyncGenerator[SystemManager, None]:
    manager = SystemManager("user", "pass", session, "pymultiMATIC", SERIAL, defaults.SENSO)
    await senso_connector.login()
    with mock.patch.object(senso_connector, "request", wraps=senso_connector.request):
        manager._connector = senso_connector
        yield manager


@pytest_asyncio.fixture(name="managers")
async def fixture_managers(
    session: ClientSession,
    connector: Connector,
    manager: SystemManager,
    senso_manager: SystemManager,
) -> AsyncGenerator[List[SystemManager], None]:
    yield [manager, senso_manager]


@pytest.mark.asyncio
async def test_system(manager: SystemManager, resp: aioresponses) -> None:
    with open(path("files/responses/livereport"), "r") as file:
        livereport_data = json.loads(file.read())

    with open(path("files/responses/rooms"), "r") as file:
        rooms_data = json.loads(file.read())

    with open(path("files/responses/systemcontrol"), "r") as file:
        system_data = json.loads(file.read())

    with open(path("files/responses/hvacstate"), "r") as file:
        hvacstate_data = json.loads(file.read())

    with open(path("files/responses/facilities"), "r") as file:
        facilities = json.loads(file.read())

    with open(path("files/responses/gateway"), "r") as file:
        gateway = json.loads(file.read())

    _mock_urls(
        resp,
        hvacstate_data,
        livereport_data,
        rooms_data,
        system_data,
        facilities,
        gateway,
    )

    system = await manager.get_system()

    assert system is not None

    assert len(system.zones) == 2
    assert len(system.rooms) == 4
    _assert_calls(6, manager)
    assert manager._fixed_serial


@pytest.mark.asyncio
async def test_system_senso_vr921(senso_manager: SystemManager, resp: aioresponses) -> None:
    with open(path("files/responses/senso/vr921/live_report"), "r") as file:
        livereport_data = json.loads(file.read())

    with open(path("files/responses/senso/vr921/rooms"), "r") as file:
        rooms_data = json.loads(file.read())

    with open(path("files/responses/senso/vr921/system"), "r") as file:
        system_data = json.loads(file.read())

    with open(path("files/responses/senso/vr921/hvac"), "r") as file:
        hvacstate_data = json.loads(file.read())

    with open(path("files/responses/senso/vr921/facilities_list"), "r") as file:
        facilities = json.loads(file.read())

    with open(path("files/responses/senso/vr921/gateway_type"), "r") as file:
        gateway = json.loads(file.read())

    _mock(
        urls_senso,
        resp,
        hvacstate_data,
        livereport_data,
        rooms_data,
        system_data,
        facilities,
        gateway,
    )

    system = await senso_manager.get_system()

    assert system is not None

    assert len(system.zones) == 3
    assert len(system.rooms) == 0
    # Rooms API is not called
    _assert_calls(5, senso_manager)
    assert senso_manager._fixed_serial


@pytest.mark.asyncio
async def test_system_senso_vr920(senso_manager: SystemManager, resp: aioresponses) -> None:
    with open(path("files/responses/senso/vr920/live_report"), "r") as file:
        livereport_data = json.loads(file.read())

    with open(path("files/responses/senso/vr920/rooms"), "r") as file:
        rooms_data = json.loads(file.read())

    with open(path("files/responses/senso/vr920/system"), "r") as file:
        system_data = json.loads(file.read())

    with open(path("files/responses/senso/vr920/hvac"), "r") as file:
        hvacstate_data = json.loads(file.read())

    with open(path("files/responses/senso/vr920/facilities_list"), "r") as file:
        facilities = json.loads(file.read())

    with open(path("files/responses/senso/vr920/gateway_type"), "r") as file:
        gateway = json.loads(file.read())

    _mock(
        urls_senso,
        resp,
        hvacstate_data,
        livereport_data,
        rooms_data,
        system_data,
        facilities,
        gateway,
    )

    system = await senso_manager.get_system()

    assert system is not None

    assert len(system.zones) == 1
    assert len(system.rooms) == 7
    _assert_calls(6, senso_manager)
    assert senso_manager._fixed_serial


@pytest.mark.asyncio
async def test_get_hot_water(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        with open(path("files/responses/hotwater"), "r") as file:
            raw_hotwater = json.loads(file.read())

        dhw_url = manager.urls.hot_water(id="Control_DHW", serial=SERIAL)
        resp.get(dhw_url, payload=raw_hotwater, status=200)

        hot_water = await manager.get_hot_water("Control_DHW")

        assert hot_water is not None
        _assert_calls(1, manager, [dhw_url])


@pytest.mark.asyncio
async def test_set_hot_water_setpoint_temperature(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.hot_water_temperature_setpoint(id="id", serial=SERIAL)
        payload = manager.payloads.hotwater_temperature_setpoint(60.0)

        resp.put(url, status=200)

        await manager.set_hot_water_setpoint_temperature("id", 60)

        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_set_hot_water_setpoint_temp_number_to_round(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.hot_water_temperature_setpoint(serial=SERIAL, id="id")
        payload = manager.payloads.hotwater_temperature_setpoint(60.5)

        resp.put(url, status=200)

        await manager.set_hot_water_setpoint_temperature("id", 60.4)
        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_set_quick_mode_no_current_quick_mode(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.system_quickmode(serial=SERIAL)
        payload = manager.payloads.quickmode(QuickModes.VENTILATION_BOOST.name)

        resp.put(url, status=200)

        await manager.set_quick_mode(QuickModes.VENTILATION_BOOST)
        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_logout(manager: SystemManager) -> None:
    await manager.logout()
    _assert_calls(1, manager, [manager.urls.logout()])


# logout tests need to be run individually (loss of cookie)
@pytest.mark.asyncio
async def test_logout_senso(senso_manager: SystemManager) -> None:
    await test_logout(senso_manager)


@pytest.mark.asyncio
async def test_set_quick_veto_room(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.room_quick_veto(serial=SERIAL, id="1")

        quick_veto = QuickVeto(100, 25)
        resp.put(url, status=200)

        await manager.set_room_quick_veto("1", quick_veto)
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_set_hot_water_operation_mode_wrong_mode(managers: List[SystemManager]) -> None:
    for manager in managers:
        await manager.set_hot_water_operating_mode("hotwater", OperatingModes.NIGHT)

        _assert_calls(0, manager)


@pytest.mark.asyncio
async def test_set_hot_water_operation_mode_heating_mode(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.hot_water_operating_mode(serial=SERIAL, id="hotwater")

        resp.put(url, status=200)
        await manager.set_hot_water_operating_mode("hotwater", OperatingModes.ON)
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_set_quick_veto_zone(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.zone_quick_veto(id="Zone1", serial=SERIAL)

        quick_veto = QuickVeto(duration=100, target=25)
        resp.put(url, status=200)

        await manager.set_zone_quick_veto("Zone1", quick_veto)
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_set_room_operation_mode_heating_mode(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.room_operating_mode(id="1", serial=SERIAL)
        print(url)

        resp.put(url, status=200)
        await manager.set_room_operating_mode("1", OperatingModes.AUTO)
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_set_room_operation_mode_no_new_mode(managers: List[SystemManager]) -> None:
    for manager in managers:
        await manager.set_room_operating_mode("1", None)
        _assert_calls(0, manager)


@pytest.mark.asyncio
async def test_set_room_operation_mode_wrong_mode(managers: List[SystemManager]) -> None:
    for manager in managers:
        await manager.set_room_operating_mode("1", OperatingModes.NIGHT)


@pytest.mark.asyncio
async def test_set_zone_operation_mode_heating_mode(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.zone_heating_mode(id="Zone1", serial=SERIAL)

        resp.put(url, status=200)
        await manager.set_zone_heating_operating_mode("Zone1", OperatingModes.AUTO)
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_set_zone_operation_mode_no_new_mode(managers: List[SystemManager]) -> None:
    for manager in managers:
        await manager.set_zone_heating_operating_mode("Zone1", None)
        _assert_calls(0, manager)


@pytest.mark.asyncio
async def test_set_zone_operation_mode_no_zone(managers: List[SystemManager]) -> None:
    for manager in managers:
        await manager.set_zone_heating_operating_mode(None, OperatingModes.MANUAL)
        _assert_calls(0, manager)


@pytest.mark.asyncio
async def test_get_room(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        with open(path("files/responses/room"), "r") as file:
            raw_rooms = json.loads(file.read())

        resp.get(manager.urls.room(id="1", serial=SERIAL), payload=raw_rooms, status=200)

        new_room = await manager.get_room("1")
        assert new_room is not None


@pytest.mark.asyncio
async def test_get_zone(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        with open(path("files/responses/zone"), "r") as file:
            raw_zone = json.loads(file.read())

        url = manager.urls.zone(serial=SERIAL, id="Control_ZO2")
        resp.get(url, payload=raw_zone, status=200)

        new_zone = await manager.get_zone("Control_ZO2")
        assert new_zone is not None
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_circulation(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        with open(path("files/responses/circulation"), "r") as file:
            raw_circulation = json.loads(file.read())

        url = manager.urls.circulation(id="id_dhw", serial=SERIAL)
        resp.get(url, payload=raw_circulation, status=200)

        new_circulation = await manager.get_circulation("id_dhw")
        assert new_circulation is not None
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_set_room_setpoint_temperature(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.room_temperature_setpoint(id="1", serial=SERIAL)
        payload = manager.payloads.room_temperature_setpoint(22.0)
        resp.put(url, status=200)

        await manager.set_room_setpoint_temperature("1", 22)
        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_set_zone_setpoint_temperature(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.zone_heating_setpoint_temperature(id="Zone1", serial=SERIAL)
        payload = manager.payloads.zone_temperature_setpoint(25.5)

        resp.put(url, status=200)

        await manager.set_zone_heating_setpoint_temperature("Zone1", 25.5)
        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_set_zone_setback_temperature(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.zone_heating_setback_temperature(id="Zone1", serial=SERIAL)
        payload = manager.payloads.zone_temperature_setback(18.0)

        resp.put(url, status=200)

        await manager.set_zone_heating_setback_temperature("Zone1", 18)
        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_set_holiday_mode(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        tomorrow = date.today() + timedelta(days=1)
        after_tomorrow = tomorrow + timedelta(days=1)

        url = manager.urls.system_holiday_mode(serial=SERIAL)
        resp.put(url, status=200)
        payload = manager.payloads.holiday_mode(True, tomorrow, after_tomorrow, 15.0)

        await manager.set_holiday_mode(tomorrow, after_tomorrow, 15)
        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_remove_holiday_mode(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        yesterday = date.today() - timedelta(days=1)
        before_yesterday = yesterday - timedelta(days=1)

        url = manager.urls.system_holiday_mode(serial=SERIAL)
        resp.put(url, status=200)
        payload = manager.payloads.holiday_mode(
            False, before_yesterday, yesterday, constants.FROST_PROTECTION_TEMP
        )

        await manager.remove_holiday_mode()
        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_remove_zone_quick_veto(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.zone_quick_veto(id="id", serial=SERIAL)
        resp.delete(url, status=200)

        await manager.remove_zone_quick_veto("id")
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_remove_room_quick_veto(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.room_quick_veto(id="1", serial=SERIAL)
        resp.delete(url, status=200)

        await manager.remove_room_quick_veto("1")
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_request_hvac_update(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url_update = manager.urls.hvac_update(serial=SERIAL)
        resp.put(url_update, status=200)

        with open(path("files/responses/hvacstate"), "r") as file:
            hvacstate_data = json.loads(file.read())

        url_hvac = manager.urls.hvac(serial=SERIAL)
        resp.get(url_hvac, payload=hvacstate_data, status=200)

        await manager.request_hvac_update()

        _assert_calls(2, manager, [url_hvac, url_update])


@pytest.mark.asyncio
async def test_request_hvac_not_sync(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url_update = manager.urls.hvac_update(serial=SERIAL)
        resp.put(url_update, status=200)

        with open(path("files/responses/hvacstate_pending"), "r") as file:
            hvacstate_data = json.loads(file.read())

        url_hvac = manager.urls.hvac(serial=SERIAL)
        resp.get(url_hvac, payload=hvacstate_data, status=200)

        await manager.request_hvac_update()
        _assert_calls(1, manager, [url_hvac])


@pytest.mark.asyncio
async def test_remove_quick_mode(manager: SystemManager, resp: aioresponses) -> None:
    url = manager.urls.system_quickmode(serial=SERIAL)
    resp.delete(url, status=200)

    await manager.remove_quick_mode()
    _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_remove_quick_mode_senso(senso_manager: SystemManager, resp: aioresponses) -> None:
    url = senso_manager.urls.system_quickmode(serial=SERIAL)
    resp.delete(url, status=200)

    await senso_manager.remove_quick_mode()
    # The QuickMode APIs are not compatible with SENSO and are therefore not called.
    calls = senso_manager._connector.request.call_args_list  # type: ignore
    assert 0 == len(calls)


@pytest.mark.asyncio
async def test_remove_quick_mode_no_active_quick_mode(
    manager: SystemManager, resp: aioresponses
) -> None:
    url = manager.urls.system_quickmode(serial=SERIAL)
    resp.delete(url, status=409)

    await manager.remove_quick_mode()
    _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_remove_quick_mode_no_active_quick_mode_senso(
    senso_manager: SystemManager, resp: aioresponses
) -> None:
    url = senso_manager.urls.system_quickmode(serial=SERIAL)
    resp.delete(url, status=409)

    await senso_manager.remove_quick_mode()
    # The QuickMode APIs are not compatible with SENSO and are therefore not called.
    calls = senso_manager._connector.request.call_args_list  # type: ignore
    assert 0 == len(calls)


@pytest.mark.asyncio
async def test_remove_quick_mode_error(manager: SystemManager, resp: aioresponses) -> None:
    url = manager.urls.system_quickmode(serial=SERIAL)
    resp.delete(url, status=400)

    try:
        await manager.remove_quick_mode()
        raise AssertionError("Error expected")
    except ApiError as exc:
        assert exc.status == 400

    _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_remove_quick_mode_error_senso(
    senso_manager: SystemManager, resp: aioresponses
) -> None:
    url = senso_manager.urls.system_quickmode(serial=SERIAL)
    resp.delete(url, status=400)

    await senso_manager.remove_quick_mode()

    # The QuickMode APIs are not compatible with SENSO and are therefore not called.
    calls = senso_manager._connector.request.call_args_list  # type: ignore
    assert 0 == len(calls)


@pytest.mark.asyncio
async def test_quick_veto_temperature_room_rounded(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.room_quick_veto(id="0", serial=SERIAL)
        payload = manager.payloads.room_quick_veto(22.5, 180)
        resp.put(url, status=200)

        qveto = QuickVeto(180, 22.7)
        await manager.set_room_quick_veto("0", qveto)

        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_quick_veto_temperature_zone_rounded(
    manager: SystemManager, resp: aioresponses
) -> None:
    url = manager.urls.zone_quick_veto(id="zone1", serial=SERIAL)
    payload = manager.payloads.zone_quick_veto(22.5)
    resp.put(url, status=200)

    qveto = QuickVeto(duration=35, target=22.7)
    await manager.set_zone_quick_veto("zone1", qveto)

    _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_quick_veto_temperature_zone_rounded_senso(
    senso_manager: SystemManager, resp: aioresponses
) -> None:
    url = senso_manager.urls.zone_quick_veto(id="zone1", serial=SERIAL)
    payload = senso_manager.payloads.zone_quick_veto(22.5, 35)
    resp.put(url, status=200)

    qveto = QuickVeto(duration=35, target=22.7)
    await senso_manager.set_zone_quick_veto("zone1", qveto)

    _assert_calls(1, senso_manager, [url], [payload])


@pytest.mark.asyncio
async def test_holiday_mode_temperature_rounded(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.system_holiday_mode(serial=SERIAL)
        resp.put(url, status=200)

        tomorrow = date.today() + timedelta(days=1)
        after_tomorrow = tomorrow + timedelta(days=1)

        payload = manager.payloads.holiday_mode(True, tomorrow, after_tomorrow, 22.5)

        await manager.set_holiday_mode(tomorrow, after_tomorrow, 22.7)

        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_serial_not_fixed(session: ClientSession) -> None:
    manager = SystemManager("user", "pass", session, "pymultiMATIC")
    assert not manager._fixed_serial


@pytest.mark.asyncio
async def test_serial_not_fixed_login(session: ClientSession, resp: aioresponses) -> None:
    manager = SystemManager("user", "pass", session, "pymultiMATIC")

    with open(path("files/responses/zone"), "r") as file:
        raw_zone = json.loads(file.read())

    url = manager.urls.zone(serial=SERIAL, id="zone")
    resp.get(url, payload=raw_zone, status=200)

    await manager.get_zone("zone")
    assert manager._serial == SERIAL
    assert not manager._fixed_serial


@pytest.mark.asyncio
async def test_serial_not_fixed_relogin(
    session: ClientSession, connector: Connector, resp: aioresponses
) -> None:
    manager = SystemManager("user", "pass", session, "pymultiMATIC")

    with open(path("files/responses/zone"), "r") as file:
        raw_zone = json.loads(file.read())

    with open(path("files/responses/facilities"), "r") as file:
        facilities = json.loads(file.read())

    facilities["body"]["facilitiesList"][0]["serialNumber"] = "123"

    url_zone1 = manager.urls.zone(serial=SERIAL, id="zone")
    url_zone2 = manager.urls.zone(serial="123", id="zone")

    url_facilities = manager.urls.facilities_list(serial=SERIAL)

    resp.get(url_zone1, payload=raw_zone, status=200)
    resp.get(url_zone2, payload=raw_zone, status=200)
    resp.get(url_facilities, payload=facilities, status=200)

    mock_auth(resp)

    await manager.get_zone("zone")
    assert manager._serial == SERIAL
    assert not manager._fixed_serial

    connector._clear_cookies()

    await manager.get_zone("zone")
    assert manager._serial == "123"


@pytest.mark.asyncio
async def test_login(session: ClientSession) -> None:
    manager = SystemManager("user", "pass", session, "pymultiMATIC")
    assert await manager.login()


@pytest.mark.asyncio
async def test_logout_serial_not_fixed(session: ClientSession) -> None:
    manager = SystemManager("user", "pass", session, "pymultiMATIC")
    assert await manager.login()
    await manager.logout()
    assert manager._serial is None


@pytest.mark.asyncio
async def test_set_ventilation_operating_mode(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.set_ventilation_operating_mode(
            serial=SERIAL,
            id="123",
        )
        resp.put(url, status=200)

        payload = manager.payloads.ventilation_operating_mode("OFF")

        await manager.set_ventilation_operating_mode("123", OperatingModes.OFF)

        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_get_gateway(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.gateway_type(
            serial=SERIAL,
        )
        with open(path("files/responses/gateway"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        gateway = await manager.get_gateway()
        assert gateway == "VR920"
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_quickmode(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.system_quickmode(
            serial=SERIAL,
        )
        with open(path("files/responses/quick_mode"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        quickmode = await manager.get_quick_mode()
        assert quickmode == QuickModes.SYSTEM_OFF
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_quickmode_no_quickmode(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.system_quickmode(
            serial=SERIAL,
        )

        resp.get(url, status=409)

        quickmode = await manager.get_quick_mode()
        assert quickmode is None
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_outdoor_temperature(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.system_status(
            serial=SERIAL,
        )

        with open(path("files/responses/systemstatus"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        temp = await manager.get_outdoor_temperature()
        assert temp == 12.5
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_hvac_status(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.hvac(
            serial=SERIAL,
        )

        with open(path("files/responses/hvacstate"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        await manager.get_hvac_status()
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_facility_detail_no_serial(
    managers: List[SystemManager], resp: aioresponses
) -> None:
    for manager in managers:
        url = manager.urls.facilities_list(
            serial=SERIAL,
        )

        with open(path("files/responses/facilities"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        details = await manager.get_facility_detail()
        assert details.serial_number == SERIAL
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_facility_detail_other_serial(manager: SystemManager, resp: aioresponses) -> None:
    url = manager.urls.facilities_list(
        serial=SERIAL,
    )

    with open(path("files/responses/facilities_multiple"), "r") as file:
        json_raw = json.loads(file.read())

    key = None
    for match in resp._matches.items():
        if match[1].url_or_pattern.path in manager.urls.facilities_list():  # type: ignore
            key = match[0]
    resp._matches.pop(key)
    resp.get(url, status=200, payload=json_raw)

    details = await manager.get_facility_detail("888")
    assert details.serial_number == "888"
    _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_facility_detail_other_serial_senso(
    senso_manager: SystemManager, resp: aioresponses
) -> None:
    url = senso_manager.urls.facilities_list(
        serial=SERIAL,
    )

    with open(path("files/responses/facilities_multiple"), "r") as file:
        json_raw = json.loads(file.read())

    key = None
    for match in resp._matches.items():
        if match[1].url_or_pattern.path in senso_manager.urls.facilities_list():  # type: ignore
            key = match[0]
    resp._matches.pop(key)
    resp.get(url, status=200, payload=json_raw)

    details = await senso_manager.get_facility_detail("888")
    assert details.serial_number == "888"
    _assert_calls(1, senso_manager, [url])


@pytest.mark.asyncio
async def test_get_live_reports(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.live_report(
            serial=SERIAL,
        )

        with open(path("files/responses/livereport"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        reports = await manager.get_live_reports()
        assert reports is not None and len(reports) > 0
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_live_report(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.live_report_device(serial=SERIAL, report_id="1", device_id="2")

        with open(path("files/responses/livereport_single"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        reports = await manager.get_live_report("1", "2")
        assert reports is not None
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_holiday_mode(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.system_holiday_mode(
            serial=SERIAL,
        )

        with open(path("files/responses/holiday_mode"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        await manager.get_holiday_mode()
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_rooms(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.rooms(
            serial=SERIAL,
        )

        with open(path("files/responses/rooms"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        rooms = await manager.get_rooms()
        assert rooms is not None and len(rooms) > 0
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_dhw(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.dhws(
            serial=SERIAL,
        )

        with open(path("files/responses/dhws"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        dhw = await manager.get_dhw()
        assert dhw.hotwater is not None and dhw.circulation is not None
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_zones(manager: SystemManager, resp: aioresponses) -> None:
    url = urls.zones(
        serial=SERIAL,
    )

    with open(path("files/responses/zones"), "r") as file:
        json_raw = json.loads(file.read())

    resp.get(url, status=200, payload=json_raw)

    zones = await manager.get_zones()
    assert zones is not None and len(zones) > 0
    _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_zones_senso_vr920(senso_manager: SystemManager, resp: aioresponses) -> None:
    url = urls_senso.zones(
        serial=SERIAL,
    )

    with open(path("files/responses/senso/vr920/zones"), "r") as file:
        json_raw = json.loads(file.read())

    resp.get(url, status=200, payload=json_raw)

    zones = await senso_manager.get_zones()
    assert zones is not None and len(zones) > 0
    _assert_calls(1, senso_manager, [url])


@pytest.mark.asyncio
async def test_get_zones_senso_vr921(senso_manager: SystemManager, resp: aioresponses) -> None:
    url = urls_senso.zones(
        serial=SERIAL,
    )

    with open(path("files/responses/senso/vr921/zones"), "r") as file:
        json_raw = json.loads(file.read())

    resp.get(url, status=200, payload=json_raw)

    zones = await senso_manager.get_zones()
    assert zones is not None and len(zones) > 0
    _assert_calls(1, senso_manager, [url])


@pytest.mark.asyncio
async def test_get_ventilation(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.system_ventilation(
            serial=SERIAL,
        )

        with open(path("files/responses/ventilation"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        ventilation = await manager.get_ventilation()
        assert ventilation is not None
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_get_emf_devices(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.emf_devices(
            serial=SERIAL,
        )

        with open(path("files/responses/emf_devices"), "r") as file:
            json_raw = json.loads(file.read())

        resp.get(url, status=200, payload=json_raw)

        emf_reports = await manager.get_emf_devices()
        assert emf_reports is not None
        assert len(emf_reports) == 7
        _assert_calls(1, manager, [url])


@pytest.mark.asyncio
async def test_setdatetime(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.system_datetime(
            serial=SERIAL,
        )

        resp.put(url, status=200)

        dt = datetime.datetime.now()
        payload = {"datetime": dt.isoformat(timespec="microseconds")}
        await manager.set_datetime(dt)
        _assert_calls(1, manager, [url], [payload])


@pytest.mark.asyncio
async def test_setdatetime_no_micro(managers: List[SystemManager], resp: aioresponses) -> None:
    for manager in managers:
        url = manager.urls.system_datetime(
            serial=SERIAL,
        )

        resp.put(url, status=200)

        dt = datetime.datetime.strptime(
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), "%Y-%m-%d %H:%M:%S"
        )
        payload = {"datetime": dt.isoformat(timespec="microseconds")}
        await manager.set_datetime(dt)
        _assert_calls(1, manager, [url], [payload])


def _mock(
    urls_class: Any,
    resp: aioresponses,
    hvacstate_data: Any,
    livereport_data: Any,
    rooms_data: Any,
    system_data: Any,
    facilities: Any = None,
    gateway: Any = None,
) -> None:
    resp.get(urls_class.live_report(serial=SERIAL), payload=livereport_data, status=200)
    resp.get(urls_class.rooms(serial=SERIAL), payload=rooms_data, status=200)
    resp.get(urls_class.system(serial=SERIAL), payload=system_data, status=200)
    resp.get(urls_class.hvac(serial=SERIAL), payload=hvacstate_data, status=200)

    if facilities:
        resp.get(urls_class.facilities_list(), payload=facilities, status=200)

    if gateway:
        resp.get(urls_class.gateway_type(serial=SERIAL), payload=gateway, status=200)


def _mock_urls(
    resp: aioresponses,
    hvacstate_data: Any,
    livereport_data: Any,
    rooms_data: Any,
    system_data: Any,
    facilities: Any = None,
    gateway: Any = None,
) -> None:
    _mock(
        urls,
        resp,
        hvacstate_data,
        livereport_data,
        rooms_data,
        system_data,
        facilities,
        gateway,
    )


def _assert_calls(
    count: int,
    manager: SystemManager,
    expected_urls: List[str] = None,
    expected_payloads: List[Any] = None,
) -> None:
    calls = manager._connector.request.call_args_list  # type: ignore
    assert count == len(calls)

    actual_urls: List[str] = []
    actual_payloads: List[Dict[str, Any]] = []

    for call in calls:
        (args, kwargs) = call
        actual_urls.append(args[1])
        actual_payloads.append(args[2])

    if expected_urls:
        diff = [x for x in expected_urls if x not in actual_urls]
        assert not diff

    if expected_payloads:
        diff = [x for x in expected_payloads if x not in actual_payloads]
        assert not diff


def _api_error(status: int) -> ApiError:
    return ApiError(message="api error", response="blah", status=status)


def _wrong_response_error(status: int) -> ApiError:
    return WrongResponseError(message="api error", response="blah", status=status)


@pytest.mark.parametrize(
    "on_exceptions, on_status_codes, exception, should_retry, expect_ex",
    [
        ((ValueError,), (), ValueError(), True, _api_error(500)),
        ((ValueError,), (), IndexError(), False, IndexError()),
        ((ValueError,), (500,), IndexError(), False, IndexError()),
        ((ValueError,), (), _api_error(400), False, _api_error(400)),
        ((ValueError,), (500,), _api_error(400), False, _api_error(400)),
        (
            (WrongResponseError,),
            (200,),
            _wrong_response_error(200),
            True,
            _api_error(200),
        ),
    ],
)
@pytest.mark.asyncio
async def test_retry_async(
    on_exceptions: Tuple[Type[BaseException]],
    on_status_codes: Tuple[int],
    exception: Type[BaseException],
    should_retry: bool,
    expect_ex: Type[BaseException],
) -> None:
    cnt = {"cnt": 0}
    num_tries = 3

    @retry_async(
        num_tries=num_tries,
        on_exceptions=on_exceptions,
        on_status_codes=on_status_codes,
        backoff_base=0,
    )
    async def func() -> None:
        cnt["cnt"] += 1
        raise exception

    with pytest.raises(expect_ex.__class__):  # type: ignore
        await func()

    assert cnt["cnt"] == (num_tries if should_retry else 1)



"""Groups everything related to the ventilation."""
import attr

from . import ActiveMode, Component, Function, OperatingModes


@attr.s
class Ventilation(Function, Component):
    """Represent the ventilation."""

    MODES = [OperatingModes.OFF, OperatingModes.NIGHT, OperatingModes.DAY]
    """List of mode that are applicable to ventilation."""

    MAX_LEVEL: float = 6
    """Maximum level for ventilation"""
    MIN_LEVEL: float = 1
    """Minimum level for ventilation"""

    temperature = attr.ib(default=None, init=False)

    def _active_mode(self) -> ActiveMode:
        if self.operating_mode == OperatingModes.OFF:
            mode = ActiveMode(Ventilation.MIN_LEVEL, OperatingModes.OFF)
        elif self.operating_mode == OperatingModes.DAY:
            mode = ActiveMode(self.target_high, OperatingModes.DAY)
        else:  # MODE_NIGHT
            mode = ActiveMode(self.target_low, OperatingModes.NIGHT)
        return mode



"""Test for rooms."""
import unittest

from pymultimatic.model import OperatingModes, Room
from tests.conftest import _room


class RoomTest(unittest.TestCase):
    """Test class."""

    def test_get_active_mode_manual(self) -> None:
        """Test active mode manual."""
        room = _room()
        room.operating_mode = OperatingModes.MANUAL
        room.target_high = 5

        active_mode = room.active_mode

        self.assertEqual(OperatingModes.MANUAL, active_mode.current)
        self.assertEqual(5, active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_off(self) -> None:
        """Test active mode off."""
        room = _room()
        room.operating_mode = OperatingModes.OFF

        active_mode = room.active_mode

        self.assertEqual(OperatingModes.OFF, active_mode.current)
        self.assertEqual(Room.MIN_TARGET_TEMP, active_mode.target)
        self.assertIsNone(active_mode.sub)



"""Groups all time program related functionality. Time program is used when
:class:`~pymultimatic.model.mode.OperatingModes.AUTO` operation mode is
activated."""
import copy
import re
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

import attr

from . import SettingMode


def _to_absolute_minutes(start_time: str) -> int:
    """Convert hh:mm to absolute minutes."""
    split = start_time.split(":")
    hour = int(split[0]) * 60
    minute = int(split[1])
    return hour + minute


def _to_time(minutes: int) -> str:
    """Convert absolute minutes to hh:mm."""
    hour = minutes // 60
    minute = minutes % 60
    return f"{hour:02n}:{minute:02n}"


@attr.s
class TimePeriodSetting:
    """This is a period setting, defining what the
    :class:`~pymultimatic.model.component.Component` should do when on
    :class:`~pymultimatic.model.mode.OperatingModes.AUTO`.

    There is not end of the period, the end of the period is equal to the start
    time of the next period.

    Args:
        start_time (str): Start time of the period, format is HH:mm.
        target_temperature (float): Target temperature to reach during this
        time period, target temperature is only available for
            :class:`~pymultimatic.model.component.Room`.
        setting (SettingMode): The setting configured, see
            :class:`~pymultimatic.model.mode.SettingModes`, this is not
            available for a :class:`~pymultimatic.model.component.Room`.
    """

    start_time = attr.ib(type=str)
    target_temperature = attr.ib(type=Optional[float])
    setting = attr.ib(type=Optional[SettingMode])
    hour = attr.ib(type=int, init=False)
    """Hour at which the setting is starting."""
    minute = attr.ib(type=int, init=False)
    """Minute at which the setting is starting."""
    absolute_minutes = attr.ib(type=int, init=False)
    """Represents :attr:`start_time` in absolute minute. hours * 60 + minutes.
    This is more convenient to compare TimePeriodSetting using this."""
    end_time = attr.ib(type=str, default=None)

    def __attrs_post_init__(self) -> None:
        self.absolute_minutes = _to_absolute_minutes(self.start_time)
        self.hour = int(self.start_time.split(":")[0])
        self.minute = int(self.start_time.split(":")[1])

    @staticmethod
    def _validate_time(value: Any) -> None:
        validator = re.compile("[0-9]{1,2}:[0-9]{2}")
        if not validator.match(value):
            raise ValueError(value)

    @start_time.validator
    def _validate_start_time(self, attribute: Any, value: Any) -> None:
        self._validate_time(value)

    @end_time.validator
    def _validate_end_time(self, attribute: Any, value: Any) -> None:
        # end time is only present on Senso timeprogram
        if value is not None:
            self._validate_time(value)

    def __deepcopy__(self, memodict: Any = None) -> "TimePeriodSetting":
        return TimePeriodSetting(
            self.start_time, self.target_temperature, self.setting, self.end_time
        )


@attr.s
class TimeProgramDay:
    """This is a day, this is basically a list of :class:`TimePeriodSetting`.

    Args:
        settings (List[TimePeriodSetting]): List of periods for this day.
    """

    settings = attr.ib(type=List[TimePeriodSetting])

    def complete_empty_periods(self, setting_mode_for_completion: SettingMode) -> None:
        """Complete the empty periods with the mode provided as a parameter.

        Args:
            setting_mode_for_completion (SettingMode): mode for empty periods
        """
        # Sorted the configuration to find the empty periods
        settings = sorted(self.settings, key=lambda period: period.absolute_minutes)
        last_end_time_absolute_minutes = 0
        to_add = []

        for setting in settings:
            if setting.absolute_minutes - last_end_time_absolute_minutes > 0:
                start = _to_time(last_end_time_absolute_minutes)
                end = _to_time(setting.absolute_minutes)
                to_add.append(TimePeriodSetting(start, None, setting_mode_for_completion, end))
            last_end_time_absolute_minutes = _to_absolute_minutes(setting.end_time)

        # End of day completion if necessary
        if last_end_time_absolute_minutes < 1440:
            start = _to_time(
                last_end_time_absolute_minutes if last_end_time_absolute_minutes > 0 else 0
            )
            to_add.append(TimePeriodSetting(start, None, setting_mode_for_completion, "24:00"))

        self.settings.extend(to_add)
        self.settings = sorted(self.settings, key=lambda period: period.absolute_minutes)


@attr.s
class TimeProgram:
    """This is the full time program, a week, reflecting the configuration done
    through mobile app.

    Args:
        days (Dict[str, TimeProgramDay]): Days of the week.
    """

    days = attr.ib(type=Dict[str, TimeProgramDay])

    def get_for(self, search_date: datetime) -> Optional[TimePeriodSetting]:
        """Get the corresponding :class:`TimePeriodSetting`
        for the given time.

        Args:
            search_date (datetime): Only the day, the hour and minute are used
                in order to get the right :class:`TimePeriodSetting`.

        Returns:
            TimePeriodSetting: The corresponding setting.
        """
        if self.days:
            day = search_date.strftime("%A").lower()
            day_before = (search_date - timedelta(days=1)).strftime("%A").lower()
            time = str(search_date.hour) + ":" + str(search_date.minute)

            abs_minutes = _to_absolute_minutes(time)
            tp_day = self.days.get(day)
            tp_day_before = self.days.get(day_before)

            if tp_day and tp_day_before:
                # if given hour:minute is before the first setting of the day,
                # get last setting of the previous day
                if abs_minutes < tp_day.settings[0].absolute_minutes:
                    return copy.deepcopy(tp_day_before.settings[-1])

                idx: int = 0
                max_len: int = len(tp_day.settings)
                while idx < max_len and abs_minutes > tp_day.settings[idx].absolute_minutes:
                    idx += 1

                if not idx == max_len:
                    # At this point, we went 1 step too far, so idx - 1
                    return copy.deepcopy(tp_day.settings[idx - 1])

                # if no match a this point, it means search date is after the last
                # setting of the day
                return copy.deepcopy(tp_day.settings[-1])
        return None

    def get_next(self, search_date: datetime) -> TimePeriodSetting:
        """
        Get the next :class:`TimePeriodSetting` for the given time.
        So if there is a time program like this:

        * 5:00

        * 6:00

        And It's 5:30, it will return the second one

        Args:
            search_date (datetime): Only the day, the hour and minute are used
                in order to get the right :class:`TimePeriodSetting`.

        Returns:
            TimePeriodSetting: The corresponding setting.
        """
        day = search_date.strftime("%A").lower()
        day_after = (search_date + timedelta(days=1)).strftime("%A").lower()
        time = str(search_date.hour) + ":" + str(search_date.minute)

        abs_minutes = _to_absolute_minutes(time)
        tp_day = self.days[day]
        tp_day_after = self.days[day_after]

        # if given hour:minute is before the first setting of the day,
        # get the first setting of the day
        if abs_minutes < tp_day.settings[0].absolute_minutes:
            return copy.deepcopy(tp_day.settings[0])

        idx: int = 0
        max_len: int = len(tp_day.settings)
        while idx < max_len and abs_minutes > tp_day.settings[idx].absolute_minutes:
            idx += 1

        if not idx == max_len:
            # At this point, we are on next setting, so it's fine
            return copy.deepcopy(tp_day.settings[idx])

        # if no match a this point, it means search date is after the last
        # setting of the day, so next one, is the first setting of the next day
        return copy.deepcopy(tp_day_after.settings[0])



"""Test for holiday mode."""
import unittest
from datetime import date, timedelta

from pymultimatic.model import HolidayMode


class HolidayModeTest(unittest.TestCase):
    """Test class."""

    def test_is_active_false(self) -> None:
        """Test non active."""
        mode = HolidayMode(False, None, None, None)
        self.assertFalse(mode.is_applied)

    def test_is_active_active_no_dates(self) -> None:
        """Test active without dates."""
        mode = HolidayMode(True, None, None, None)
        self.assertFalse(mode.is_applied)

    def test_is_active_active_not_between(self) -> None:
        """Test active today not between start and end dates."""
        today = date.today()
        start_date = today - timedelta(days=today.weekday() - 2)
        end_date = today - timedelta(days=today.weekday() - 1)
        mode = HolidayMode(True, start_date, end_date, 15)
        self.assertFalse(mode.is_applied)

    def test_is_active_active_between(self) -> None:
        """Test active today between start and end dates."""
        today = date.today()
        start_date = today - timedelta(days=1)
        end_date = today + timedelta(days=1)
        mode = HolidayMode(True, start_date, end_date, 15)
        self.assertTrue(mode.is_applied)



import unittest

from pymultimatic.model import OperatingModes, SettingModes
from tests.conftest import _full_day_time_program, _ventilation


class ZoneTest(unittest.TestCase):
    def test_get_active_mode_night(self) -> None:
        ventilation = _ventilation()
        ventilation.operating_mode = OperatingModes.NIGHT
        active_mode = ventilation.active_mode

        self.assertEqual(OperatingModes.NIGHT, active_mode.current)
        self.assertEqual(ventilation.target_low, active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_day(self) -> None:
        ventilation = _ventilation()
        ventilation.operating_mode = OperatingModes.DAY
        active_mode = ventilation.active_mode

        self.assertEqual(OperatingModes.DAY, active_mode.current)
        self.assertEqual(ventilation.target_high, active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_off(self) -> None:
        ventilation = _ventilation()
        ventilation.operating_mode = OperatingModes.OFF

        active_mode = ventilation.active_mode

        self.assertEqual(OperatingModes.OFF, active_mode.current)
        self.assertEqual(1, active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_auto_on(self) -> None:
        ventilation = _ventilation()
        ventilation.operating_mode = OperatingModes.AUTO

        active_mode = ventilation.active_mode

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(3, active_mode.target)
        self.assertEqual(SettingModes.ON, active_mode.sub)

    def test_get_active_mode_auto_off(self) -> None:
        ventilation = _ventilation()
        ventilation.time_program = _full_day_time_program(SettingModes.OFF)
        ventilation.operating_mode = OperatingModes.AUTO

        active_mode = ventilation.active_mode

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(1, active_mode.target)
        self.assertEqual(SettingModes.OFF, active_mode.sub)



# pylint: disable=all
"""Vaillant API Urls for multimatic app."""
from typing import Any
from urllib import parse

_BASE = "https://smart.vaillant.com/mobile/api/v4"

_BASE_AUTHENTICATE = _BASE + "/account/authentication/v1"
_AUTHENTICATE = _BASE_AUTHENTICATE + "/authenticate"
_NEW_TOKEN = _BASE_AUTHENTICATE + "/token/new"
_LOGOUT = _BASE_AUTHENTICATE + "/logout"

"""Facility details"""
_FACILITIES_LIST = _BASE + "/facilities"
_FACILITIES = _FACILITIES_LIST + "/{serial}"
_GATEWAY_TYPE = _FACILITIES + "/public/v1/gatewayType"
_FACILITIES_DETAILS = _FACILITIES + "/system/v1/details"
_FACILITIES_STATUS = _FACILITIES + "/system/v1/status"
_FACILITIES_SETTINGS = _FACILITIES + "/storage"
_FACILITIES_DEFAULT_SETTINGS = _FACILITIES + "/storage/default"
_FACILITIES_INSTALLER_INFO = _FACILITIES + "/system/v1/installerinfo"

"""Rbr (Room by room)"""
_RBR_BASE = _FACILITIES + "/rbr/v1"
_RBR_INSTALLATION_STATUS = _RBR_BASE + "/installationStatus"
_RBR_UNDERFLOOR_HEATING_STATUS = _RBR_BASE + "/underfloorHeatingStatus"

"""Rooms"""
_ROOMS_LIST = _RBR_BASE + "/rooms"
_ROOM = _ROOMS_LIST + "/{id}"
_ROOM_CONFIGURATION = _ROOM + "/configuration"
_ROOM_QUICK_VETO = _ROOM_CONFIGURATION + "/quickVeto"
_ROOM_TIMEPROGRAM = _ROOM + "/timeprogram"
_ROOM_OPERATING_MODE = _ROOM_CONFIGURATION + "/operationMode"
_ROOM_CHILD_LOCK = _ROOM_CONFIGURATION + "/childLock"
_ROOM_NAME = _ROOM_CONFIGURATION + "/name"
_ROOM_DEVICE_NAME = _ROOM_CONFIGURATION + "/devices/{sgtin}/name"
_ROOM_TEMPERATURE_SETPOINT = _ROOM_CONFIGURATION + "/temperatureSetpoint"

"""Repeaters"""
_REPEATERS_LIST = _RBR_BASE + "/repeaters"
_REPEATER_DELETE = _REPEATERS_LIST + "/{sgtin}"
_REPEATER_SET_NAME = _REPEATERS_LIST + "/{sgtin}/name"

"""HVAC (heating, ventilation and Air-conditioning)"""
_HVAC = _FACILITIES + "/hvacstate/v1/overview"
_HVAC_REQUEST_UPDATE = _FACILITIES + "/hvacstate/v1/hvacMessages/update"

"""EMF (Embedded Metering Function)"""
_LIVE_REPORT = _FACILITIES + "/livereport/v1"
_LIVE_REPORT_DEVICE = _LIVE_REPORT + "/devices/{device_id}/reports/{report_id}"
_PHOTOVOLTAICS_REPORT = _FACILITIES + "/spine/v1/currentPVMeteringInfo"
_EMF_DEVICES = _FACILITIES + "/emf/v1/devices"
_EMF_REPORT_DEVICE = _EMF_DEVICES + "/{device_id}"

"""System control"""
_SYSTEM = _FACILITIES + "/systemcontrol/v1"
_SYSTEM_CONFIGURATION = _SYSTEM + "/configuration"
_SYSTEM_STATUS = _SYSTEM + "/status"
_SYSTEM_DATETIME = _SYSTEM_STATUS + "/datetime"
_SYSTEM_PARAMETERS = _SYSTEM + "/parameters"
_SYSTEM_QUICK_MODE = _SYSTEM_CONFIGURATION + "/quickmode"
_SYSTEM_HOLIDAY_MODE = _SYSTEM_CONFIGURATION + "/holidaymode"

"""DHW (Domestic Hot Water)"""
_DHWS = _SYSTEM + "/dhw"
_DHW = _DHWS + "/{id}"

"""Circulation"""
_CIRCULATION = _DHW + "/circulation"
_CIRCULATION_CONFIGURATION = _CIRCULATION + "/configuration"
_CIRCULATION_TIMEPROGRAM = _CIRCULATION_CONFIGURATION + "/timeprogram"

"""Hot water"""
_HOT_WATER = _DHW + "/hotwater"
_HOT_WATER_CONFIGURATION = _HOT_WATER + "/configuration"
_HOT_WATER_TIMEPROGRAM = _HOT_WATER_CONFIGURATION + "/timeprogram"
_HOT_WATER_OPERATING_MODE = _HOT_WATER_CONFIGURATION + "/operation_mode"
_HOT_WATER_TEMPERATURE_SETPOINT = _HOT_WATER_CONFIGURATION + "/temperature_setpoint"

"""Ventilation"""
_SYSTEM_VENTILATION = _SYSTEM + "/ventilation"
_VENTILATION = _SYSTEM + "/ventilation/{id}"
_VENTILATION_CONFIGURATION = _VENTILATION + "/fan/configuration"
_VENTILATION_TIMEPROGRAM = _VENTILATION_CONFIGURATION + "/timeprogram"
_VENTILATION_DAY_LEVEL = _VENTILATION_CONFIGURATION + "/day_level"
_VENTILATION_NIGHT_LEVEL = _VENTILATION_CONFIGURATION + "/night_level"
_VENTILATION_OPERATING_MODE = _VENTILATION_CONFIGURATION + "/operation_mode"

"""Zones"""
_ZONES_LIST = _SYSTEM + "/zones"
_ZONE = _ZONES_LIST + "/{id}"
_ZONE_CONFIGURATION = _ZONE + "/configuration"
_ZONE_NAME = _ZONE_CONFIGURATION + "/name"
_ZONE_QUICK_VETO = _ZONE_CONFIGURATION + "/quick_veto"

"""Zone heating"""
_ZONE_HEATING_CONFIGURATION = _ZONE + "/heating/configuration"
_ZONE_HEATING_TIMEPROGRAM = _ZONE + "/heating/timeprogram"
_ZONE_HEATING_MODE = _ZONE_HEATING_CONFIGURATION + "/mode"
_ZONE_HEATING_SETPOINT_TEMPERATURE = _ZONE_HEATING_CONFIGURATION + "/setpoint_temperature"
_ZONE_HEATING_SETBACK_TEMPERATURE = _ZONE_HEATING_CONFIGURATION + "/setback_temperature"

"""Zone cooling"""
_ZONE_COOLING_CONFIGURATION = _ZONE + "/cooling/configuration"
_ZONE_COOLING_TIMEPROGRAM = _ZONE + "/cooling/timeprogram"
_ZONE_COOLING_MODE = _ZONE_COOLING_CONFIGURATION + "/mode"
_ZONE_COOLING_SETPOINT_TEMPERATURE = _ZONE_COOLING_CONFIGURATION + "/setpoint_temperature"
_ZONE_COOLING_MANUAL_SETPOINT_TEMPERATURE = (
    _ZONE_COOLING_CONFIGURATION + "/manual_mode_cooling_temperature_setpoint"
)


def base(**kwargs: Any) -> str:
    """Base url of the API."""
    return _BASE.format(**kwargs)


def new_token(**kwargs: Any) -> str:
    """Url to request a new token."""
    return _NEW_TOKEN.format(**kwargs)


def authenticate(**kwargs: Any) -> str:
    """Url to authenticate the user and receive cookies."""
    return _AUTHENTICATE.format(**kwargs)


def logout(**kwargs: Any) -> str:
    """Url to logout from the API, cookies are invalidated."""
    return _LOGOUT.format(**kwargs)


def facilities_list(**kwargs: Any) -> str:
    """Url to get the list of serial numbers of the facilities (and some other
    properties).

    Note:
        For now, the connector only handle one serial number.
    """
    return _FACILITIES_LIST.format(**kwargs)


def gateway_type(**kwargs: Any) -> str:
    """Url to get the gateway type (VR900, VR920, etc.)."""
    return _GATEWAY_TYPE.format(**kwargs)


def rbr_underfloor_heating_status(**kwargs: Any) -> str:
    """Url to check if underfloor heating is installed or not."""
    return _RBR_UNDERFLOOR_HEATING_STATUS.format(**kwargs)


def rbr_installation_status(**kwargs: Any) -> str:
    """Url to check the room by room installation status."""
    return _RBR_INSTALLATION_STATUS.format(**kwargs)


def rooms(**kwargs: Any) -> str:
    """Url to get the list of :class:`~pymultimatic.model.component.Room`."""
    return _ROOMS_LIST.format(**kwargs)


def room(**kwargs: Any) -> str:
    """Url to get specific room details (configuration, timeprogram). Or to
    delete a :class:`~pymultimatic.model.component.Room`.
    """
    return _ROOM.format(**kwargs)


def room_configuration(**kwargs: Any) -> str:
    """Url to get configuration for a
    :class:`~pymultimatic.model.component.Room` (name, temperature,
    target temperature, etc.).
    """
    return _ROOM_CONFIGURATION.format(**kwargs)


def room_quick_veto(**kwargs: Any) -> str:
    """Url to handle :class:`~pymultimatic.model.mode.QuickVeto` for a
    :class:`~pymultimatic.model.component.Room`.
    """
    return _ROOM_QUICK_VETO.format(**kwargs)


def room_operating_mode(**kwargs: Any) -> str:
    """Url to set operating for a :class:`~pymultimatic.model.component.Room`."""
    return _ROOM_OPERATING_MODE.format(**kwargs)


def room_timeprogram(**kwargs: Any) -> str:
    """Url to get/update configuration for a
    class:`~pymultimatic.model.component.Room`. (name, temperature,
    target temperature, etc.).
    """
    return _ROOM_TIMEPROGRAM.format(**kwargs)


def room_child_lock(**kwargs: Any) -> str:
    """Url to handle child lock for all
    :class:`~pymultimatic.model.component.Device` in a
    :class:`~pymultimatic.model.component.Room`.
    """
    return _ROOM_CHILD_LOCK.format(**kwargs)


def room_name(**kwargs: Any) -> str:
    """Set :class:`~pymultimatic.model.component.Room` name."""
    return _ROOM_NAME.format(**kwargs)


def room_device_name(**kwargs: Any) -> str:
    """Set :class:`~pymultimatic.model.component.Device` name."""
    return _ROOM_DEVICE_NAME.format(**kwargs)


def room_temperature_setpoint(**kwargs: Any) -> str:
    """Url to handle target temperature for a
    :class:`~pymultimatic.model.component.Room`.
    """
    return _ROOM_TEMPERATURE_SETPOINT.format(**kwargs)


def repeaters(**kwargs: Any) -> str:
    """Url to get list of repeaters"""
    return _REPEATERS_LIST.format(**kwargs)


def delete_repeater(**kwargs: Any) -> str:
    """Url to delete a repeater."""
    return _REPEATER_DELETE.format(**kwargs)


def repeater_name(**kwargs: Any) -> str:
    """Url to set repeater's name."""
    return _REPEATER_SET_NAME.format(**kwargs)


def hvac(**kwargs: Any) -> str:
    """Url of the hvac overview."""
    return _HVAC.format(**kwargs)


def hvac_update(**kwargs: Any) -> str:
    """Url to request an hvac update."""
    return _HVAC_REQUEST_UPDATE.format(**kwargs)


def live_report(**kwargs: Any) -> str:
    """Url to get live report data (current boiler water temperature, current
    hot water temperature, etc.)."""
    return _LIVE_REPORT.format(**kwargs)


def live_report_device(**kwargs: Any) -> str:
    """
    Url to get live report for specific device
    """
    return _LIVE_REPORT_DEVICE.format(**kwargs)


def photovoltaics(**kwargs: Any) -> str:
    """Url to get photovoltaics data."""
    return _PHOTOVOLTAICS_REPORT.format(**kwargs)


def emf_devices(**kwargs: Any) -> str:
    """Url to get emf (Embedded Metering Function) report."""
    return _EMF_DEVICES.format(**kwargs)


def emf_report_device(
    energy_type: str, function: str, time_range: str, start: str, offset: str, **kwargs: Any
) -> str:
    """Url to get emf (Embedded Metering Function) report for a specific
    device."""
    url = _EMF_REPORT_DEVICE.format(**kwargs)

    query_params = {
        "energyType": energy_type,
        "function": function,
        "timeRange": time_range,
        "start": start,
        "offset": offset,
    }

    return "{}?{}".format(url, parse.urlencode(query_params))


def facilities_details(**kwargs: Any) -> str:
    """Url to get facility detail."""
    return _FACILITIES_DETAILS.format(**kwargs)


def facilities_status(**kwargs: Any) -> str:
    """Url to get facility status."""
    return _FACILITIES_STATUS.format(**kwargs)


def facilities_settings(**kwargs: Any) -> str:
    """Url to get facility settings."""
    return _FACILITIES_SETTINGS.format(**kwargs)


def facilities_default_settings(**kwargs: Any) -> str:
    """
    Url to get facility default settings
    """
    return _FACILITIES_DEFAULT_SETTINGS.format(**kwargs)


def facilities_installer_info(**kwargs: Any) -> str:
    """Url to get facility default settings."""
    return _FACILITIES_INSTALLER_INFO.format(**kwargs)


def system(**kwargs: Any) -> str:
    """Url to get full :class:`~pymultimatic.model.system.System` (zones, dhw,
    ventilation, holiday mode, etc.) except
    :class:`~pymultimatic.model.component.Room`.
    """
    return _SYSTEM.format(**kwargs)


def system_configuration(**kwargs: Any) -> str:
    """Url to get system configuration (holiday mode, quick mode etc.)."""
    return _SYSTEM_CONFIGURATION.format(**kwargs)


def system_status(**kwargs: Any) -> str:
    """Url to get outdoor temperature and datetime."""
    return _SYSTEM_STATUS.format(**kwargs)


def system_datetime(**kwargs: Any) -> str:
    """Url to set datetime."""
    return _SYSTEM_DATETIME.format(**kwargs)


def system_parameters(**kwargs: Any) -> str:
    """Url to get system parameters."""
    return _SYSTEM_PARAMETERS.format(**kwargs)


def system_quickmode(**kwargs: Any) -> str:
    """Url to get system :class:`~pymultimatic.model.mode.QuickMode`."""
    return _SYSTEM_QUICK_MODE.format(**kwargs)


def system_holiday_mode(**kwargs: Any) -> str:
    """Url to get system :class:`~pymultimatic.model.mode.HolidayMode`."""
    return _SYSTEM_HOLIDAY_MODE.format(**kwargs)


def dhw(**kwargs: Any) -> str:
    """Url to get domestic hot water
    (:class:`~pymultimatic.model.component.HotWater` and
    :class:`~pymultimatic.model.component.Circulation`).
    """
    return _DHW.format(**kwargs)


def dhws(**kwargs: Any) -> str:
    """Url to get all domestic hot water
    (:class:`~pymultimatic.model.component.HotWater` and
    :class:`~pymultimatic.model.component.Circulation`).
    """
    return _DHWS.format(**kwargs)


def circulation(**kwargs: Any) -> str:
    """Url to get :class:`~pymultimatic.model.component.Circulation` details."""
    return _CIRCULATION.format(**kwargs)


def circulation_configuration(**kwargs: Any) -> str:
    """Url to handle :class:`~pymultimatic.model.component.Circulation`
    configuration.
    """
    return _CIRCULATION_CONFIGURATION.format(**kwargs)


def circulation_timeprogram(**kwargs: Any) -> str:
    """Url to handle :class:`~pymultimatic.model.component.Circulation`
    :class:`~pymultimatic.model.timeprogram.TimeProgram`.
    """
    return _CIRCULATION_TIMEPROGRAM.format(**kwargs)


def hot_water(**kwargs: Any) -> str:
    """Url to get :class:`~pymultimatic.model.component.HotWater` detail."""
    return _HOT_WATER.format(**kwargs)


def hot_water_configuration(**kwargs: Any) -> str:
    """Url to handle :class:`~pymultimatic.model.component.HotWater`
    configuration.
    """
    return _HOT_WATER_CONFIGURATION.format(**kwargs)


def hot_water_timeprogram(**kwargs: Any) -> str:
    """Url to handle :class:`~pymultimatic.model.component.HotWater`
    :class:`~pymultimatic.model.timeprogram.TimeProgram`.
    """
    return _HOT_WATER_TIMEPROGRAM.format(**kwargs)


def hot_water_operating_mode(**kwargs: Any) -> str:
    """Url to set :class:`~pymultimatic.model.component.HotWater`
    operating mode, only if it's not a quick action.
    """
    return _HOT_WATER_OPERATING_MODE.format(**kwargs)


def hot_water_temperature_setpoint(**kwargs: Any) -> str:
    """Url to set :class:`~pymultimatic.model.component.HotWater`
    temperature setpoint.
    """
    return _HOT_WATER_TEMPERATURE_SETPOINT.format(**kwargs)


def ventilation(**kwargs: Any) -> str:
    """Url to get ventilation details."""
    return _VENTILATION.format(**kwargs)


def system_ventilation(**kwargs: Any) -> str:
    """Url to get ventilation details."""
    return _SYSTEM_VENTILATION.format(**kwargs)


def ventilation_configuration(**kwargs: Any) -> str:
    """Url to get ventilation configuration."""
    return _VENTILATION_CONFIGURATION.format(**kwargs)


def ventilation_timeprogram(**kwargs: Any) -> str:
    """Url to get ventilation timeprogram."""
    return _VENTILATION_TIMEPROGRAM.format(**kwargs)


def set_ventilation_day_level(**kwargs: Any) -> str:
    """Url to set ventilation day level."""
    return _VENTILATION_DAY_LEVEL.format(**kwargs)


def set_ventilation_night_level(**kwargs: Any) -> str:
    """
    Url to set ventilation night level
    """
    return _VENTILATION_NIGHT_LEVEL.format(**kwargs)


def set_ventilation_operating_mode(**kwargs: Any) -> str:
    """Url to set ventilation operating mode."""
    return _VENTILATION_OPERATING_MODE.format(**kwargs)


def zones(**kwargs: Any) -> str:
    """Url to get :class:`~pymultimatic.model.component.Zone`."""
    return _ZONES_LIST.format(**kwargs)


def zone(**kwargs: Any) -> str:
    """Url to get a specific :class:`~pymultimatic.model.component.Zone`."""
    return _ZONE.format(**kwargs)


def zone_configuration(**kwargs: Any) -> str:
    """Url to get a specific :class:`~pymultimatic.model.component.Zone`
    configuration.
    """
    return _ZONE_CONFIGURATION.format(**kwargs)


def zone_name(**kwargs: Any) -> str:
    """Url to set :class:`~pymultimatic.model.component.Zone` name."""
    return _ZONE_NAME.format(**kwargs)


def zone_quick_veto(**kwargs: Any) -> str:
    """Url to get :class:`~pymultimatic.model.mode.QuickVeto` for a
    :class:`~pymultimatic.model.component.Zone`.
    """
    return _ZONE_QUICK_VETO.format(**kwargs)


def zone_heating_configuration(**kwargs: Any) -> str:
    """Url to get :class:`~pymultimatic.model.component.Zone` heating
    configuration.
    """
    return _ZONE_HEATING_CONFIGURATION.format(**kwargs)


def zone_heating_timeprogram(**kwargs: Any) -> str:
    """Url to get a :class:`~pymultimatic.model.component.Zone` heating
    :class:`~pymultimatic.model.timeprogram.TimeProgram`.
    """
    return _ZONE_HEATING_TIMEPROGRAM.format(**kwargs)


def zone_heating_mode(**kwargs: Any) -> str:
    """Url to get a :class:`~pymultimatic.model.component.Zone` heating mode."""
    return _ZONE_HEATING_MODE.format(**kwargs)


def zone_heating_setpoint_temperature(**kwargs: Any) -> str:
    """Url to set a :class:`~pymultimatic.model.component.Zone` setpoint
    temperature.
    """
    return _ZONE_HEATING_SETPOINT_TEMPERATURE.format(**kwargs)


def zone_heating_setback_temperature(**kwargs: Any) -> str:
    """Url to set a :class:`~pymultimatic.model.component.Zone` setback
    temperature.
    """
    return _ZONE_HEATING_SETBACK_TEMPERATURE.format(**kwargs)


def zone_cooling_configuration(**kwargs: Any) -> str:
    """Url to get a :class:`~pymultimatic.model.component.Zone` cooling
    configuration.
    """
    return _ZONE_COOLING_CONFIGURATION.format(**kwargs)


def zone_cooling_timeprogram(**kwargs: Any) -> str:
    """Url to get :class:`~pymultimatic.model.component.Zone` cooling
    timeprogram.
    """
    return _ZONE_COOLING_TIMEPROGRAM.format(**kwargs)


def zone_cooling_mode(**kwargs: Any) -> str:
    """Url to set a :class:`~pymultimatic.model.component.Zone` cooling mode."""
    return _ZONE_COOLING_MODE.format(**kwargs)


def zone_cooling_setpoint_temperature(**kwargs: Any) -> str:
    """Url to set the cooling temperature setpoint for a
    :class:`~pymultimatic.model.component.Zone`.
    """
    return _ZONE_COOLING_SETPOINT_TEMPERATURE.format(**kwargs)


def zone_cooling_manual_setpoint_temperature(**kwargs: Any) -> str:
    """Url to set manual cooling setpoint temperature for a
    :class:`~pymultimatic.model.component.Zone`.
    """
    return _ZONE_COOLING_MANUAL_SETPOINT_TEMPERATURE.format(**kwargs)



import unittest

from pymultimatic.model import ActiveFunction, OperatingModes, SettingModes, Zone, ZoneCooling
from tests.conftest import _full_day_time_program, _zone, _zone_cooling


class ZoneTest(unittest.TestCase):
    def test_get_active_mode_night(self) -> None:
        zone = _zone()
        zone.heating.operating_mode = OperatingModes.NIGHT
        active_mode = zone.active_mode

        self.assertEqual(OperatingModes.NIGHT, active_mode.current)
        self.assertEqual(zone.heating.target_low, active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_day(self) -> None:
        zone = _zone()
        zone.heating.operating_mode = OperatingModes.DAY
        active_mode = zone.active_mode

        self.assertEqual(OperatingModes.DAY, active_mode.current)
        self.assertEqual(zone.heating.target_high, active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_off(self) -> None:
        zone = _zone()
        zone.heating.operating_mode = OperatingModes.OFF

        active_mode = zone.active_mode

        self.assertEqual(OperatingModes.OFF, active_mode.current)
        self.assertEqual(Zone.MIN_TARGET_HEATING_TEMP, active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_cooling_active_mode_auto(self) -> None:
        cooling = ZoneCooling()
        cooling.operating_mode = OperatingModes.AUTO
        cooling.time_program = _full_day_time_program()

        active_mode = cooling.active_mode
        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.ON, active_mode.sub)

    def test_cooling_active_mode_auto_off(self) -> None:
        cooling = ZoneCooling()
        cooling.operating_mode = OperatingModes.AUTO
        cooling.time_program = _full_day_time_program(mode=SettingModes.OFF)

        active_mode = cooling.active_mode
        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.OFF, active_mode.sub)

    def test_cooling_active_mode_on(self) -> None:
        cooling = ZoneCooling()
        cooling.operating_mode = OperatingModes.ON
        cooling.time_program = _full_day_time_program()

        active_mode = cooling.active_mode
        self.assertEqual(OperatingModes.ON, active_mode.current)
        self.assertIsNone(active_mode.sub)

    def test_cooling_active_mode_off(self) -> None:
        cooling = ZoneCooling()
        cooling.operating_mode = OperatingModes.OFF
        cooling.time_program = _full_day_time_program()

        active_mode = cooling.active_mode
        self.assertEqual(OperatingModes.OFF, active_mode.current)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_cooling_auto(self) -> None:
        zone = _zone_cooling()
        active_mode = zone.active_mode

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(zone.cooling.target_high, active_mode.target)
        self.assertEqual(SettingModes.ON, active_mode.sub)

    def test_get_active_mode_standby_cooling(self) -> None:
        zone = _zone_cooling()
        zone.active_function = ActiveFunction.STANDBY
        active_mode = zone.active_mode

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(zone.cooling.target_high, active_mode.target)
        self.assertEqual(SettingModes.ON, active_mode.sub)

    def test_get_active_mode_standby_heating(self) -> None:
        zone = _zone()
        zone.active_function = ActiveFunction.STANDBY
        active_mode = zone.active_mode

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(zone.heating.target_high, active_mode.target)
        self.assertEqual(SettingModes.DAY, active_mode.sub)



"""Test for system."""
import datetime
import unittest

from pymultimatic.model import (
    Dhw,
    HolidayMode,
    HotWater,
    OperatingModes,
    QuickModes,
    QuickVeto,
    Room,
    SettingModes,
    System,
    TimePeriodSetting,
    TimeProgramDay,
    Ventilation,
    Zone,
    constants,
    TimeProgram,
)
from tests.conftest import (
    _circulation,
    _hotwater,
    _room,
    _full_day_time_program,
    _zone,
    _zone_senso,
    _zone_cooling,
)


class SystemTest(unittest.TestCase):
    """Test class."""

    def test_get_active_mode_room(self) -> None:
        """Test active mode room."""
        room = _room()

        system = System(rooms=[room])

        active_mode = system.get_active_mode_room(room)

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertIsNone(active_mode.sub)
        self.assertEqual(20, active_mode.target)

    def test_get_active_mode_room_quick_veto(self) -> None:
        """Test mode room with quick veto."""
        room = _room()
        quick_veto = QuickVeto(duration=0, target=22)
        room.quick_veto = quick_veto
        system = System(rooms=[room])

        active_mode = system.get_active_mode_room(room)

        self.assertEqual(OperatingModes.QUICK_VETO, active_mode.current)
        self.assertEqual(quick_veto.target, active_mode.target)

    def test_get_active_mode_room_holiday_mode(self) -> None:
        """Test active mode room with holiday mode."""
        holiday = HolidayMode(True, datetime.date.today(), datetime.date.today(), 10)

        room = _room()
        system = System(rooms=[room], holiday=holiday)

        active_mode = system.get_active_mode_room(room)

        self.assertEqual(QuickModes.HOLIDAY, active_mode.current)
        self.assertEqual(holiday.target, active_mode.target)

    def test_get_active_mode_room_system_off(self) -> None:
        """Test active mode room when system off."""
        room = _room()
        system = System(rooms=[room], quick_mode=QuickModes.SYSTEM_OFF)

        active_mode = system.get_active_mode_room(room)

        self.assertEqual(QuickModes.SYSTEM_OFF, active_mode.current)
        self.assertEqual(Room.MIN_TARGET_TEMP, active_mode.target)

    def test_get_active_mode_hot_water(self) -> None:
        """Test get active mode for hot water."""

        dhw = Dhw(hotwater=_hotwater())
        system = System(dhw=dhw)

        active_mode = system.get_active_mode_hot_water()

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.ON, active_mode.sub)
        self.assertEqual(50, active_mode.target)

    def test_get_active_mode_hot_water_no_hotwater(self) -> None:
        """Test active mode without hot water."""
        system = System(quick_mode=QuickModes.HOTWATER_BOOST)

        active_mode = system.get_active_mode_hot_water()
        self.assertIsNone(active_mode)

    def test_get_active_mode_hot_water_off(self) -> None:
        """Test active mode hot water off."""
        hotwater = _hotwater()
        hotwater.time_program = _full_day_time_program(SettingModes.OFF)
        dhw = Dhw(hotwater=hotwater)

        system = System(dhw=dhw)

        active_mode = system.get_active_mode_hot_water()

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.OFF, active_mode.sub)
        self.assertEqual(HotWater.MIN_TARGET_TEMP, active_mode.target)

    def test_get_active_mode_hot_water_party(self) -> None:
        """Test active mode hot water off."""
        hotwater = _hotwater()
        hotwater.time_program = _full_day_time_program(SettingModes.OFF)
        dhw = Dhw(hotwater=hotwater)

        system = System(dhw=dhw, quick_mode=QuickModes.PARTY)

        active_mode = system.get_active_mode_hot_water()

        self.assertEqual(QuickModes.PARTY, active_mode.current)
        self.assertEqual(hotwater.target_high, active_mode.target)

    def test_get_active_mode_hot_water_system_off(self) -> None:
        """Test active mode hot water system off."""
        dhw = Dhw(hotwater=_hotwater())
        system = System(dhw=dhw, quick_mode=QuickModes.SYSTEM_OFF)

        active_mode = system.get_active_mode_hot_water()

        self.assertEqual(QuickModes.SYSTEM_OFF, active_mode.current)
        self.assertEqual(constants.FROST_PROTECTION_TEMP, active_mode.target)

    def test_get_active_mode_hot_water_one_day_away(self) -> None:
        """Test get active mode for hot water with one day away."""
        dhw = Dhw(hotwater=_hotwater())
        system = System(dhw=dhw, quick_mode=QuickModes.ONE_DAY_AWAY)

        active_mode = system.get_active_mode_hot_water()

        self.assertEqual(QuickModes.ONE_DAY_AWAY, active_mode.current)
        self.assertEqual(constants.FROST_PROTECTION_TEMP, active_mode.target)

    def test_get_active_mode_hot_water_hot_water_boost(self) -> None:
        """Test get active mode for hot water with hot water boost."""
        temp = 55

        hot_water = _hotwater()
        hot_water.operating_mode = OperatingModes.ON
        hot_water.target_high = temp
        dhw = Dhw(hotwater=hot_water)
        system = System(dhw=dhw, quick_mode=QuickModes.HOTWATER_BOOST)

        active_mode = system.get_active_mode_hot_water()

        self.assertEqual(active_mode.current, QuickModes.HOTWATER_BOOST)
        self.assertEqual(temp, active_mode.target)

    def test_get_active_mode_hot_water_holiday_mode(self) -> None:
        """Test get active mode for hot water with holiday mode."""
        holiday = HolidayMode(True, datetime.date.today(), datetime.date.today(), 10)

        dhw = Dhw(hotwater=_hotwater())

        system = System(holiday=holiday, dhw=dhw)

        active_mode = system.get_active_mode_hot_water()

        self.assertEqual(QuickModes.HOLIDAY, active_mode.current)
        self.assertEqual(constants.FROST_PROTECTION_TEMP, active_mode.target)

    def test_get_active_mode_zone(self) -> None:
        """Test get active mode for zone."""

        zone = _zone()
        system = System(zones=[zone])

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.DAY, active_mode.sub)
        self.assertEqual(zone.heating.target_high, active_mode.target)

    def test_get_active_mode_senso_zone(self) -> None:
        """Test get active mode for senso zone."""

        zone = _zone_senso()
        system = System(zones=[zone])

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(OperatingModes.TIME_CONTROLLED, active_mode.current)
        self.assertEqual(SettingModes.DAY, active_mode.sub)
        self.assertEqual(
            25,
            active_mode.target,
        )

    def test_get_active_mode_senso_night_zone(self) -> None:
        """Test get active mode for senso zone."""

        zone = _zone_senso(False)
        system = System(zones=[zone])

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(OperatingModes.TIME_CONTROLLED, active_mode.current)
        self.assertEqual(SettingModes.NIGHT, active_mode.sub)
        self.assertEqual(
            zone.heating.target_low,
            active_mode.target,
        )

    def test_get_active_mode_zone_off(self) -> None:
        """Test get active mode for zone off."""
        zone = _zone()
        zone.heating.time_program = _full_day_time_program(SettingModes.NIGHT)

        system = System(zones=[zone])

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.NIGHT, active_mode.sub)
        self.assertEqual(zone.heating.target_low, active_mode.target)

    def test_get_active_mode_senso_zone_off(self) -> None:
        """Test get active mode for zone off."""
        zone = _zone_senso()
        zone.heating.time_program = _full_day_time_program(SettingModes.NIGHT)

        system = System(zones=[zone])

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(OperatingModes.TIME_CONTROLLED, active_mode.current)
        self.assertEqual(SettingModes.NIGHT, active_mode.sub)
        self.assertEqual(zone.heating.target_low, active_mode.target)

    def test_get_active_mode_manual_senso_zone(self) -> None:
        """Test get active mode for zone with manual mode."""

        zone = _zone_senso()
        zone.heating.operating_mode = OperatingModes.MANUAL
        zone.heating.target_high = 25
        system = System(zones=[zone])

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(OperatingModes.MANUAL, active_mode.current)
        self.assertIsNone(active_mode.sub)
        self.assertEqual(zone.heating.target_high, active_mode.target)

    def test_get_active_mode_zone_quick_veto(self) -> None:
        """Test get active mode for zone quick veto."""
        quickveto = QuickVeto(target=30)

        zone = _zone()
        zone.quick_veto = quickveto
        system = System(zones=[zone])

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(OperatingModes.QUICK_VETO, active_mode.current)
        self.assertEqual(quickveto.target, active_mode.target)

    def test_get_active_mode_zone_holiday_mode(self) -> None:
        """Test get active mode for zone with holiday mode."""
        holiday = HolidayMode(True, datetime.date.today(), datetime.date.today(), 10)

        zone = _zone()
        system = System(holiday=holiday, zones=[zone])

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(QuickModes.HOLIDAY, active_mode.current)
        self.assertEqual(holiday.target, active_mode.target)

    def test_get_active_mode_zone_quick_mode_water_boost(self) -> None:
        """Test get active mode for zone with hot water boost."""
        zone = _zone()
        system = System(zones=[zone], quick_mode=QuickModes.HOTWATER_BOOST)

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.DAY, active_mode.sub)
        self.assertEqual(zone.heating.target_high, active_mode.target)

    def test_get_active_mode_zone_quick_mode_system_off(self) -> None:
        """Test get active mode for zone with system off."""
        zone = _zone()
        system = System(zones=[zone], quick_mode=QuickModes.SYSTEM_OFF)

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(QuickModes.SYSTEM_OFF, active_mode.current)
        self.assertEqual(Zone.MIN_TARGET_HEATING_TEMP, active_mode.target)

    def test_get_active_mode_zone_quick_mode_one_day_home(self) -> None:
        """Test get active mode for zone one day home."""

        timeprogram_day_setting_sunday = TimePeriodSetting("00:00", None, SettingModes.NIGHT)

        timeprogram = _full_day_time_program(SettingModes.DAY, None)
        timeprogram.days["sunday"] = TimeProgramDay([timeprogram_day_setting_sunday])

        zone = _zone()
        zone.heating.time_program = timeprogram
        system = System(zones=[zone], quick_mode=QuickModes.ONE_DAY_AT_HOME)

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(QuickModes.ONE_DAY_AT_HOME, active_mode.current)
        self.assertEqual(zone.heating.target_low, active_mode.target)

    def test_get_active_mode_zone_quick_mode_one_day_home_day(self) -> None:
        """Test get active mode for zone one day home."""

        timeprogram_day_setting_sunday = TimePeriodSetting("00:00", None, SettingModes.DAY)

        timeprogram = _full_day_time_program(SettingModes.NIGHT, None)
        timeprogram.days["sunday"] = TimeProgramDay([timeprogram_day_setting_sunday])

        zone = _zone()
        zone.heating.time_program = timeprogram
        system = System(zones=[zone], quick_mode=QuickModes.ONE_DAY_AT_HOME)

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(QuickModes.ONE_DAY_AT_HOME, active_mode.current)
        self.assertEqual(zone.heating.target_high, active_mode.target)

    def test_get_active_mode_zone_quick_mode_one_day_away(self) -> None:
        """Test get active mode for zone one day away."""
        zone = _zone()
        system = System(zones=[zone], quick_mode=QuickModes.ONE_DAY_AWAY)

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(QuickModes.ONE_DAY_AWAY, active_mode.current)
        self.assertEqual(Zone.MIN_TARGET_HEATING_TEMP, active_mode.target)

    def test_get_active_mode_zone_quick_mode_party(self) -> None:
        """Test get active mode for zone quick mode party."""
        zone = _zone()
        system = System(zones=[zone], quick_mode=QuickModes.PARTY)

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(QuickModes.PARTY, active_mode.current)
        self.assertEqual(zone.heating.target_high, active_mode.target)

    def test_get_active_mode_zone_quick_mode_quick_veto(self) -> None:
        """Test get active mode for zone quick mode + quick veto."""
        quick_veto = QuickVeto(target=15)
        zone = _zone()
        zone.quick_veto = quick_veto

        system = System(zones=[zone])

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(OperatingModes.QUICK_VETO, active_mode.current)
        self.assertEqual(zone.quick_veto.target, active_mode.target)

    def test_get_active_mode_zone_quick_mode_ventilation(self) -> None:
        """Test get active mode for zone quick mode ventilation."""
        zone = _zone()
        system = System(zones=[zone], quick_mode=QuickModes.VENTILATION_BOOST)

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(QuickModes.VENTILATION_BOOST, active_mode.current)
        self.assertEqual(Zone.MIN_TARGET_HEATING_TEMP, active_mode.target)

    def test_get_active_mode_circulation_hot_water_boost(self) -> None:
        """Test get active mode for circulation with hotwater boost."""
        dhw = Dhw(circulation=_circulation())

        system = System(quick_mode=QuickModes.HOTWATER_BOOST, dhw=dhw)

        active_mode = system.get_active_mode_circulation()

        self.assertEqual(QuickModes.HOTWATER_BOOST, active_mode.current)
        self.assertIsNone(active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_circulation_no_circulation(self) -> None:
        """Test get active mode for circulation without circulation."""
        system = System(quick_mode=QuickModes.HOTWATER_BOOST)

        active_mode = system.get_active_mode_circulation()
        self.assertIsNone(active_mode)

    def test_get_active_mode_circulation_off(self) -> None:
        """Test active mode for circulation off."""
        dhw = Dhw(circulation=_circulation())
        system = System(quick_mode=QuickModes.SYSTEM_OFF, dhw=dhw)

        active_mode = system.get_active_mode_circulation()

        self.assertEqual(QuickModes.SYSTEM_OFF, active_mode.current)
        self.assertIsNone(active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_circulation_holiday(self) -> None:
        """Test get active mode for circulation with holiday mode."""
        dhw = Dhw(circulation=_circulation())

        holiday = HolidayMode(True, datetime.date.today(), datetime.date.today(), 10)

        system = System(holiday=holiday, dhw=dhw)

        active_mode = system.get_active_mode_circulation()

        self.assertEqual(QuickModes.HOLIDAY, active_mode.current)
        self.assertIsNone(active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_circulation_auto(self) -> None:
        """Test get active mode for circulation."""
        dhw = Dhw(circulation=_circulation())
        system = System(dhw=dhw)

        active_mode = system.get_active_mode_circulation()

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertIsNone(active_mode.target)
        self.assertEqual(SettingModes.ON, active_mode.sub)

    def test_get_active_mode_zone_cooling_for_x_days(self) -> None:
        """Test quick mode COOLING_FOR_X_DAYS."""
        zone = _zone_cooling()
        system = System(zones=[zone], quick_mode=QuickModes.COOLING_FOR_X_DAYS)

        active_mode = system.get_active_mode_zone(zone)

        self.assertEqual(QuickModes.COOLING_FOR_X_DAYS, active_mode.current)
        self.assertIsNone(active_mode.sub)
        self.assertEqual(zone.cooling.target_high, active_mode.target)

    def test_get_active_mode_ventilation_off(self) -> None:
        """Test get active mode for ventilation."""

        ventilation = Ventilation(operating_mode=OperatingModes.OFF, target_low=3)
        system = System(ventilation=ventilation)

        active_mode = system.get_active_mode_ventilation()

        self.assertEqual(OperatingModes.OFF, active_mode.current)
        self.assertEqual(1, active_mode.target)

    def test_get_active_mode_ventilation_day(self) -> None:
        """Test get active mode for ventilation."""

        ventilation = Ventilation(operating_mode=OperatingModes.DAY, target_high=5)
        system = System(ventilation=ventilation)

        active_mode = system.get_active_mode_ventilation()

        self.assertEqual(OperatingModes.DAY, active_mode.current)
        self.assertEqual(5, active_mode.target)

    def test_get_active_mode_ventilation_night(self) -> None:
        """Test get active mode for ventilation."""

        ventilation = Ventilation(operating_mode=OperatingModes.NIGHT, target_low=3)
        system = System(ventilation=ventilation)

        active_mode = system.get_active_mode_ventilation()

        self.assertEqual(OperatingModes.NIGHT, active_mode.current)
        self.assertEqual(3, active_mode.target)

    def test_get_active_mode_ventilation_quick_mode_boost(self) -> None:
        """Test get active mode for ventilation."""

        ventilation = Ventilation(operating_mode=OperatingModes.NIGHT, target_high=5)
        system = System(ventilation=ventilation, quick_mode=QuickModes.VENTILATION_BOOST)

        active_mode = system.get_active_mode_ventilation()

        self.assertEqual(QuickModes.VENTILATION_BOOST, active_mode.current)
        self.assertEqual(Ventilation.MAX_LEVEL, active_mode.target)

    def test_get_active_mode_ventilation_one_day_away(self) -> None:
        """Test get active mode for ventilation."""

        ventilation = Ventilation(operating_mode=OperatingModes.NIGHT, target_high=5)
        system = System(ventilation=ventilation, quick_mode=QuickModes.ONE_DAY_AWAY)

        active_mode = system.get_active_mode_ventilation()

        self.assertEqual(QuickModes.ONE_DAY_AWAY, active_mode.current)
        self.assertEqual(Ventilation.MIN_LEVEL, active_mode.target)

    def test_get_active_mode_ventilation_system_off(self) -> None:
        """Test get active mode for ventilation."""

        ventilation = Ventilation(operating_mode=OperatingModes.NIGHT, target_high=5)
        system = System(ventilation=ventilation, quick_mode=QuickModes.SYSTEM_OFF)

        active_mode = system.get_active_mode_ventilation()

        self.assertEqual(QuickModes.SYSTEM_OFF, active_mode.current)
        self.assertEqual(Ventilation.MIN_LEVEL, active_mode.target)

    def test_get_active_mode_ventilation_party(self) -> None:
        """Test get active mode for ventilation."""

        ventilation = Ventilation(operating_mode=OperatingModes.NIGHT, target_high=5)
        system = System(ventilation=ventilation, quick_mode=QuickModes.PARTY)

        active_mode = system.get_active_mode_ventilation()

        self.assertEqual(QuickModes.PARTY, active_mode.current)
        self.assertEqual(ventilation.target_high, active_mode.target)

    def test_get_active_mode_ventilation_holiday(self) -> None:
        """Test get active mode for ventilation."""
        holiday = HolidayMode(True, datetime.date.today(), datetime.date.today(), 10)
        ventilation = Ventilation(operating_mode=OperatingModes.NIGHT, target_high=5)
        system = System(ventilation=ventilation, holiday=holiday)

        active_mode = system.get_active_mode_ventilation()

        self.assertEqual(QuickModes.HOLIDAY, active_mode.current)
        self.assertEqual(Ventilation.MIN_LEVEL, active_mode.target)

    def test_get_active_mode_hotwater_no_timeprogram(self) -> None:
        """Test get active mode for hot water."""

        dhw = Dhw(hotwater=_hotwater())
        dhw.hotwater.time_program = None
        system = System(dhw=dhw)

        active_mode = system.get_active_mode_hot_water()

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.ON, active_mode.sub)
        self.assertEqual(50, active_mode.target)

    def test_get_active_mode_hotwater_empty_timeprogram(self) -> None:
        """Test get active mode for hot water."""

        dhw = Dhw(hotwater=_hotwater())
        dhw.hotwater.time_program = TimeProgram(days={})
        system = System(dhw=dhw)

        active_mode = system.get_active_mode_hot_water()

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.ON, active_mode.sub)
        self.assertEqual(50, active_mode.target)

    def test_get_active_mode_hotwater_no_timeprogram_on(self) -> None:
        """Test get active mode for hot water."""

        dhw = Dhw(hotwater=_hotwater())
        dhw.hotwater.time_program = None
        dhw.hotwater.operating_mode = OperatingModes.ON
        system = System(dhw=dhw)

        active_mode = system.get_active_mode_hot_water()

        self.assertEqual(OperatingModes.ON, active_mode.current)
        self.assertEqual(50, active_mode.target)

    def test_get_active_mode_circulation_no_timeprogram(self) -> None:
        """Test get active mode for circulation."""

        dhw = Dhw(circulation=_circulation())
        dhw.circulation.time_program = None
        system = System(dhw=dhw)

        active_mode = system.get_active_mode_circulation()

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.OFF, active_mode.sub)

    def test_get_active_mode_circulation_empty_timeprogram(self) -> None:
        """Test get active mode for circulation."""

        dhw = Dhw(circulation=_circulation())
        dhw.circulation.time_program = TimeProgram(days={})
        system = System(dhw=dhw)

        active_mode = system.get_active_mode_circulation()

        self.assertEqual(OperatingModes.AUTO, active_mode.current)
        self.assertEqual(SettingModes.OFF, active_mode.sub)

    def test_get_active_mode_circulation_no_timeprogram_on(self) -> None:
        """Test get active mode for circulation."""

        dhw = Dhw(circulation=_circulation())
        dhw.circulation.time_program = None
        dhw.circulation.operating_mode = OperatingModes.ON
        system = System(dhw=dhw)

        active_mode = system.get_active_mode_circulation()

        self.assertEqual(OperatingModes.ON, active_mode.current)



"""Grouping status"""
from datetime import datetime
from typing import List, Optional

import attr


@attr.s
class Error:
    """Errors coming from your system.

    Args:
        device_name (str): Name of the device from where the error is coming.
        title (str): Short description of the error.
        status_code (str): Code of the error.
        description (str): Long description of the error.
        timestamp (datetime): When errors occurred.
    """

    device_name = attr.ib(type=str)
    title = attr.ib(type=str)
    status_code = attr.ib(type=str)
    description = attr.ib(type=str)
    timestamp = attr.ib(type=datetime)


@attr.s
class BoilerStatus(Error):
    """Status of the boiler. This is sent with an error format, but in this
    case, it's more like a status.

    Args:
        hint (str): Directly coming from the API, most of the time the value is
            ``...``
    """

    hint = attr.ib(type=str)

    @property
    def is_error(self) -> bool:
        """bool: Checks if there is an error at boiler side."""
        return self.status_code is not None and (
            self.status_code.startswith("F") or self.status_code == "con"
        )


@attr.s
class HvacStatus:
    """HVAC status.

    Args:
        online (str): Indicate if the system is connected to the cloud.
        update (str): Indicate if there is available update.
        boiler_status (BoilerStatus): Status of the boiler.
        errors (List[Error]): If there are errors, you can find them here.
    """

    online = attr.ib(type=str)
    update = attr.ib(type=str)
    boiler_status = attr.ib(type=Optional[BoilerStatus], default=None)
    errors = attr.ib(type=List[Error], default=[])

    @property
    def is_online(self) -> bool:
        """bool: Checks if the system is connected to the internet."""
        return self.online == "ONLINE"

    @property
    def is_up_to_date(self) -> bool:
        """bool: Checks if the system is up to date."""
        return self.update == "UPDATE_NOT_PENDING"



"""Test for quick veto."""
import unittest

from pymultimatic.model import QuickVeto


class QuickVetoTest(unittest.TestCase):
    """Test class."""

    def test_wrong_quick_veto(self) -> None:
        """Test duration no correct."""
        self.assertRaises(ValueError, QuickVeto, 800000, 15)

    def test_quick_veto(self) -> None:
        """Test correct quick veto."""
        quickveto = QuickVeto(duration=600, target=15)
        self.assertEqual(600, quickveto.duration)
        self.assertEqual(15, quickveto.target)

    def test_quick_veto_wrong_target(self) -> None:
        """Test target no correct."""
        self.assertRaises(ValueError, QuickVeto, 45, 45)
        self.assertRaises(ValueError, QuickVeto, 45, 4)



"""Errors coming from the API."""
from typing import Any, Optional

import attr


@attr.s
class ApiError(Exception):
    """This exception is thrown when a communication error occurs with the
    vaillant API."""

    message = attr.ib(type=str)
    response = attr.ib(type=Optional[str])
    status = attr.ib(type=int)
    payload = attr.ib(type=Any, default=None)

    def __str__(self) -> str:
        return (
            f"{self.message}, status: {self.status}, response: {self.response},"
            f"payload: {self.payload}"
        )


@attr.s
class WrongResponseError(ApiError):
    """This exception is thrown when the response coming from the API is wrong."""

    status = attr.ib(default=200, type=int)



"""contains some default parameters of the application."""

SMARTPHONE_ID = "pymultiMATIC"
"""Default smartphone id."""

SENSO = "SYSTEMCONTROL_SENSO"
MULTIMATIC = "SYSTEMCONTROL_MULTIMATIC"
"""Applications"""



"""Vaillant operation modes."""
from datetime import date
from typing import List, Optional

import attr

from . import ActiveMode, Circulation, Component, HotWater, Mode, Room, Ventilation, Zone


@attr.s(frozen=True)
class QuickMode(Mode):
    """This class is a helper to check what is impacted by a quick mode.

    Quick modes are mode you can quickly set with the mobile app, most of the
    time, they have system wise impact. This is different from
    :class:`QuickVeto`, which will only impact one
    :class:`~pymultimatic.model.component.Component`.

    Args:
        name (str): Name of the quick mode.
        duration (int): Some quick mode require a duration.
        for_zone (bool): quick mode is applicable for a
            :class:`~pymultimatic.model.component.Zone`.
        for_room (bool): quick mode is applicable for a
            :class:`~pymultimatic.model.component.Room`.
        for_dhw (bool): quick mode is applicable for a
            :class:`~pymultimatic.model.component.HotWater` and
            :class:`~pymultimatic.model.component.Circulation`.
        for_ventilation (bool): quick mode is applicable for a
            :class:`~pymultimatic.model.component.Ventilation`.

    Note:
        There is a default duration for quick modes (which can't be changed)
        but the API always returns `0` as remaining duration, so there is no
        way to know when quick mode will end.

    Vaillant documentation link about quick modes:
    http://elearning.vaillant.com/vrc700/ci/en/documents/uk/infopool/Operating_instructions.pdf
    """

    for_zone = attr.ib(type=bool, default=None, eq=False)
    for_room = attr.ib(type=bool, default=None, eq=False)
    for_dhw = attr.ib(type=bool, default=None, eq=False)
    for_ventilation = attr.ib(type=bool, default=None, eq=False)
    duration = attr.ib(type=Optional[int], default=None, eq=False)

    def is_for(self, comp: Component) -> bool:
        """Check if the quick mode has impact on the given component.

        Args:
            comp (Component): The component

        Returns:
            bool: Whether the component is affected by the quick mode or not.
        """

        if isinstance(comp, (HotWater, Circulation)) and self.for_dhw:
            return True
        if isinstance(comp, Room) and self.for_room:
            return True
        if isinstance(comp, Zone) and self.for_zone:
            return True
        if isinstance(comp, Ventilation) and self.for_ventilation:
            return True
        return False


class QuickModes:
    """Groups all :class:`QuickMode` together."""

    HOTWATER_BOOST = QuickMode("QM_HOTWATER_BOOST", False, False, True, False)
    """The advanced function heats the water in the domestic hot water cylinder
    once until the desired DHW circuit temperature set is reached or until you
    cancel the advanced function early.
    The heating installation will then return to the pre-set mode
    """

    VENTILATION_BOOST = QuickMode("QM_VENTILATION_BOOST", True, False, False, True)
    """This advanced function switches the zone off for 30 minutes.
    The frost protection function is activated, and hot water generation and
    circulation remain active.
    Ventilation is activated and works at the highest ventilation level.
    The advanced function is automatically deactivated after 30 minutes or
    if you cancel the advanced function early.
    The heating installation will then return to the pre-set mode.
    """

    ONE_DAY_AWAY = QuickMode("QM_ONE_DAY_AWAY", True, False, True, True)
    """Hot water generation and circulation are switched off and the frost
    protection is activated.
    The advanced function is automatically deactivated after 24:00 hours or if
    you cancel the advanced function first.
    The heating installation will then return to the pre-set mode.
    Ventilation is activated and works at the lowest ventilation level.
    """

    SYSTEM_OFF = QuickMode("QM_SYSTEM_OFF", True, True, True, True)
    """The heating function, hot water circuit and cooling are switched off.
    The frost protection function is activated.
    The circulation is switched off.
    Ventilation is activated and works at the lowest ventilation level.
    """

    ONE_DAY_AT_HOME = QuickMode("QM_ONE_DAY_AT_HOME", True, False, False, False)
    """This advanced function activates Automatic mode for one day with the
    settings for Sunday, as set using the Time programmes function.
    The advanced function is automatically deactivated after 24:00 hours or if
    you cancel the advanced function first.
    The heating installation will then return to the pre-set mode.
    The operating modes and time programme for
    :class:`~pymultimatic.model.component.Room` are not affected.
    """

    PARTY = QuickMode("QM_PARTY", True, False, True, True)
    """The advanced function brings the room temperature to the set desired
    Day temperature, in accordance with the set time periods.
    The advanced function is deactivated after six hours or if you cancel it
    before the six hours is up.
    The heating installation will then return to the pre-set mode.
    """

    HOLIDAY = QuickMode("QM_HOLIDAY", True, True, True, True)
    """
    """

    QUICK_VETO = QuickMode("QM_QUICK_VETO", False, False, False, False)
    """
    This advanced function activates a quick veto for one specific zone
    """

    COOLING_FOR_X_DAYS = QuickMode("QM_COOLING_FOR_X_DAYS", True, False, False, False)

    _VALUES = {
        qm.name: qm
        for qm in [
            HOTWATER_BOOST,
            VENTILATION_BOOST,
            ONE_DAY_AWAY,
            SYSTEM_OFF,
            ONE_DAY_AT_HOME,
            PARTY,
            HOLIDAY,
            QUICK_VETO,
            COOLING_FOR_X_DAYS,
        ]
    }

    @classmethod
    def for_zone(cls) -> List[QuickMode]:
        """Get the list of :class:`QuickMode` applicable to
        :class:`~pymultimatic.model.component.Zone`.

        Returns:
            A list of quick mode applicable for
            :class:`~pymultimatic.model.component.Zone`.
        """
        sub_list = []

        for quick_mode in cls._VALUES.values():
            if quick_mode.for_zone:
                sub_list.append(quick_mode)

        return sub_list

    @classmethod
    def for_room(cls) -> List[QuickMode]:
        """Get the list of :class:`QuickMode` applicable to
        :class:`~pymultimatic.model.component.Room`.

        Returns:
            A list of quick mode applicable for
            :class:`~pymultimatic.model.component.Room`.
        """
        sub_list = []

        for quick_mode in cls._VALUES.values():
            if quick_mode.for_room:
                sub_list.append(quick_mode)

        return sub_list

    @classmethod
    def for_dhw(cls) -> List[QuickMode]:
        """Get the list of :class:`QuickMode` applicable to
        :class:`~pymultimatic.model.component.HotWater`
        and :class:`~pymultimatic.model.component.Circulation`.

        Returns:
            A list of quick mode applicable for
            :class:`~pymultimatic.model.component.HotWater` and
            :class:`~pymultimatic.model.component.Circulation`.
        """
        sub_list = []

        for quick_mode in cls._VALUES.values():
            if quick_mode.for_dhw:
                sub_list.append(quick_mode)

        return sub_list

    @classmethod
    def for_ventilation(cls) -> List[QuickMode]:
        """Get the list of :class:`QuickMode` applicable to
        :class:`~pymultimatic.model.component.Ventilation`.

        Returns:
            A list of quick mode applicable for
            :class:`~pymultimatic.model.component.Ventilation`.
        """
        sub_list = []

        for quick_mode in cls._VALUES.values():
            if quick_mode.for_ventilation:
                sub_list.append(quick_mode)

        return sub_list

    @classmethod
    def get(cls, name: str, duration: Optional[int] = None) -> QuickMode:
        """Get :class:`QuickMode` by name.

        Args:
            name (str): Name of the operating mode
            duration (int): Duration of the quick mode

        Returns:
            The corresponding operating mode or None
        """
        value: QuickMode = cls._VALUES[name]
        return QuickMode(
            value.name,
            value.for_zone,
            value.for_room,
            value.for_dhw,
            value.for_ventilation,
            duration,
        )


@attr.s
class HolidayMode:
    """Represents system's holiday mode.

    Args:
        is_active (bool): Indicates whether the user has activated the holiday
            mode or not.
        start_date (date): Start date of the holiday mode.
        end_date (date): End date of the holiday mode.
        target_temperature (float): Target temperature when the
            holiday mode is applied.

    Note:
        `is_active` doesn't necessarily means the holiday mode `is_applied`:
        the user can activate the holiday mode through the mobile app but with
        `end_date` and `start_date` in the future.
    """

    is_active = attr.ib(type=bool)
    start_date = attr.ib(type=Optional[date], default=None)
    end_date = attr.ib(type=Optional[date], default=None)
    target = attr.ib(type=Optional[float], default=None)

    @property
    def active_mode(self) -> Optional[ActiveMode]:
        """Optional[ActiveMode]: Get the :class:`ActiveMode` from holiday mode,
        if `is_applied`."""
        if self.is_applied:
            return ActiveMode(self.target, QuickModes.HOLIDAY)
        return None

    @property
    def is_applied(self) -> bool:
        """bool: Check if holiday mode is active **and** if today are between
        end_date and start_date.

        Note:
            Use this to know if the holiday mode is active and applied to the
            system.
        """
        return (
            self.is_active
            and self.start_date is not None
            and self.end_date is not None
            and self.start_date <= date.today() <= self.end_date
        )



"""Test for quick mode."""
import unittest

from pymultimatic.model import Circulation, HotWater, QuickModes, Room, Ventilation, Zone


class QuickModeTest(unittest.TestCase):
    """Test class."""

    def test_for_zone(self) -> None:
        """Test get quick mode for zone."""
        values = QuickModes.for_zone()
        self.assertEqual(7, len(values))

    def test_for_room(self) -> None:
        """Test get quick mode for zone."""
        values = QuickModes.for_room()
        self.assertEqual(2, len(values))

    def test_for_dhw(self) -> None:
        """Test get quick mode for zone."""
        values = QuickModes.for_dhw()
        self.assertEqual(5, len(values))

    def test_for_ventilation(self) -> None:
        """Test get quick mode for zone."""
        values = QuickModes.for_ventilation()
        self.assertEqual(5, len(values))

    def test_is_for(self) -> None:
        for mode in QuickModes._VALUES.values():
            self.assertEqual(mode.for_zone, mode.is_for(Zone()), mode)
            self.assertEqual(mode.for_room, mode.is_for(Room()), mode)
            self.assertEqual(mode.for_dhw, mode.is_for(Circulation()), mode)
            self.assertEqual(mode.for_dhw, mode.is_for(HotWater()), mode)
            self.assertEqual(mode.for_ventilation, mode.is_for(Ventilation()), mode)



import unittest
from datetime import datetime

from pymultimatic.model import SettingModes, TimePeriodSetting, TimeProgram, TimeProgramDay


class TimeProgramTest(unittest.TestCase):
    def test_time_program_simple(self) -> None:
        tpds1 = TimePeriodSetting("00:00", 25, SettingModes.ON)
        tpds2 = TimePeriodSetting("02:00", 20, SettingModes.OFF)

        monday = TimeProgramDay([tpds1, tpds2])

        timeprogram = TimeProgram({"monday": monday, "sunday": TimeProgramDay([])})

        current = timeprogram.get_for(datetime(2019, 2, 18, 1, 0))
        self._assert(tpds1, current)

    def test_time_program_after_last(self) -> None:
        tpds1 = TimePeriodSetting("00:00", 25, SettingModes.ON)
        tpds2 = TimePeriodSetting("02:00", 20, SettingModes.OFF)

        monday = TimeProgramDay([tpds1, tpds2])

        timeprogram = TimeProgram({"monday": monday, "sunday": TimeProgramDay([])})

        current = timeprogram.get_for(datetime(2019, 2, 18, 3, 0))
        self._assert(tpds2, current)

    def test_time_program_before_first(self) -> None:
        tpds1 = TimePeriodSetting("01:00", 25, SettingModes.ON)
        tpds2 = TimePeriodSetting("02:00", 20, SettingModes.OFF)

        tpds_sunday = TimePeriodSetting("15:00", 15, SettingModes.OFF)

        sunday = TimeProgramDay([tpds_sunday])
        monday = TimeProgramDay([tpds1, tpds2])

        timeprogram = TimeProgram({"monday": monday, "sunday": sunday})

        current = timeprogram.get_for(datetime(2019, 2, 18, 0, 30))
        self._assert(tpds_sunday, current)

    def test_wrong_start_time(self) -> None:
        self.assertRaises(ValueError, TimePeriodSetting, "xx", 25, "Test1")

    def _assert(self, expected: TimePeriodSetting, actual: TimePeriodSetting) -> None:
        self.assertEqual(expected.target_temperature, actual.target_temperature)
        self.assertEqual(expected.setting, actual.setting)
        self.assertEqual(expected.start_time, actual.start_time)
        self.assertEqual(expected.hour, actual.hour)
        self.assertEqual(expected.minute, actual.minute)
        self.assertEqual(expected.absolute_minutes, actual.absolute_minutes)

    def test_get_next_before_first(self) -> None:
        tpds1 = TimePeriodSetting("01:00", 25, SettingModes.ON)
        tpds2 = TimePeriodSetting("05:00", 20, SettingModes.OFF)
        tpds3 = TimePeriodSetting("08:00", 20, SettingModes.OFF)
        tpds_day_after = TimePeriodSetting("03:00", 20, SettingModes.OFF)

        monday = TimeProgramDay([tpds1, tpds2, tpds3])
        tuesday = TimeProgramDay([tpds_day_after])

        timeprogram = TimeProgram({"monday": monday, "tuesday": tuesday})

        next_setting = timeprogram.get_next(datetime(2019, 2, 18, 0, 30))

        self._assert(next_setting, tpds1)

    def test_get_next_between_same_day(self) -> None:
        tpds1 = TimePeriodSetting("01:00", 25, SettingModes.ON)
        tpds2 = TimePeriodSetting("05:00", 20, SettingModes.OFF)
        tpds3 = TimePeriodSetting("08:00", 20, SettingModes.OFF)
        tpds_day_after = TimePeriodSetting("03:00", 20, SettingModes.OFF)

        monday = TimeProgramDay([tpds1, tpds2, tpds3])
        tuesday = TimeProgramDay([tpds_day_after])

        timeprogram = TimeProgram({"monday": monday, "tuesday": tuesday})

        next_setting = timeprogram.get_next(datetime(2019, 2, 18, 4, 30))

        self._assert(next_setting, tpds2)

    def test_get_next_after_last(self) -> None:
        tpds1 = TimePeriodSetting("01:00", 25, SettingModes.ON)
        tpds2 = TimePeriodSetting("05:00", 20, SettingModes.OFF)
        tpds3 = TimePeriodSetting("08:00", 20, SettingModes.OFF)
        tpds_day_after = TimePeriodSetting("03:00", 20, SettingModes.OFF)

        monday = TimeProgramDay([tpds1, tpds2, tpds3])
        tuesday = TimeProgramDay([tpds_day_after])

        timeprogram = TimeProgram({"monday": monday, "tuesday": tuesday})

        next_setting = timeprogram.get_next(datetime(2019, 2, 18, 9, 30))

        self._assert(next_setting, tpds_day_after)

    def test_complete_time_program(self) -> None:
        expected1 = TimePeriodSetting(
            start_time="00:00",
            end_time="01:00",
            target_temperature=None,
            setting=SettingModes.NIGHT,
        )
        tpds1 = TimePeriodSetting(
            start_time="01:00", end_time="03:00", target_temperature=25, setting=SettingModes.DAY
        )
        expected2 = TimePeriodSetting(
            start_time="03:00",
            end_time="10:00",
            target_temperature=None,
            setting=SettingModes.NIGHT,
        )
        tpds2 = TimePeriodSetting(
            start_time="10:00", end_time="10:30", target_temperature=25, setting=SettingModes.DAY
        )
        expected3 = TimePeriodSetting(
            start_time="10:30",
            end_time="20:10",
            target_temperature=None,
            setting=SettingModes.NIGHT,
        )
        tpds3 = TimePeriodSetting(
            start_time="20:10", end_time="22:00", target_temperature=25, setting=SettingModes.DAY
        )
        expected4 = TimePeriodSetting(
            start_time="22:00",
            end_time="24:00",
            target_temperature=None,
            setting=SettingModes.NIGHT,
        )

        monday = TimeProgramDay([tpds1, tpds2, tpds3])
        monday.complete_empty_periods(SettingModes.NIGHT)
        full_night_day = TimeProgramDay([])
        full_night_day.complete_empty_periods(SettingModes.NIGHT)
        timeprogram = TimeProgram(
            {"monday": monday, "tuesday": full_night_day, "sunday": full_night_day}
        )

        self._assert(expected1, timeprogram.get_for(datetime(2019, 2, 18, 0, 30)))
        self._assert(expected1, timeprogram.get_for(datetime(2019, 2, 18, 1, 0)))
        self._assert(tpds1, timeprogram.get_for(datetime(2019, 2, 18, 1, 1)))
        self._assert(expected2, timeprogram.get_for(datetime(2019, 2, 18, 9, 30)))
        self._assert(expected3, timeprogram.get_for(datetime(2019, 2, 18, 15, 0)))
        self._assert(expected4, timeprogram.get_for(datetime(2019, 2, 18, 23, 30)))

        full_night_setting = TimePeriodSetting(
            start_time="00:00",
            end_time="24:00",
            target_temperature=None,
            setting=SettingModes.NIGHT,
        )
        self._assert(full_night_setting, timeprogram.get_for(datetime(2019, 2, 19, 13, 00)))



"""Groups everything related to a zone."""
from enum import Enum
from typing import Optional

import attr

from . import ActiveMode, Component, Function, OperatingModes, constants


class ActiveFunction(Enum):
    """Represent the active function the zone is currently running.

    It can be 'STANDBY', 'COOLING', 'HEATING'
    """

    STANDBY = "STANDBY"
    COOLING = "COOLING"
    HEATING = "HEATING"


class ZoneHeating(Function):
    """Represent the heating function of a zone."""

    MODES = [
        OperatingModes.AUTO,
        OperatingModes.OFF,
        OperatingModes.DAY,
        OperatingModes.MANUAL,
        OperatingModes.NIGHT,
        OperatingModes.NORMAL,
        OperatingModes.QUICK_VETO,
        OperatingModes.REDUCED,
        OperatingModes.TIME_CONTROLLED,
    ]

    def _active_mode(self) -> ActiveMode:
        if self.operating_mode == OperatingModes.OFF:
            mode = ActiveMode(constants.FROST_PROTECTION_TEMP, OperatingModes.OFF)
        elif self.operating_mode == OperatingModes.DAY:
            mode = ActiveMode(self.target_high, OperatingModes.DAY)
        elif self.operating_mode == OperatingModes.MANUAL:
            mode = ActiveMode(self.target_high, OperatingModes.MANUAL)
        else:  # MODE_NIGHT
            mode = ActiveMode(self.target_low, OperatingModes.NIGHT)
        return mode


class ZoneCooling(Function):
    """Represent the cooling function of a zone."""

    MODES = [OperatingModes.AUTO, OperatingModes.ON, OperatingModes.OFF]

    def _active_mode(self) -> ActiveMode:
        if self.operating_mode == OperatingModes.OFF:
            mode = ActiveMode(None, OperatingModes.OFF)
        else:  # MODE ON
            mode = ActiveMode(self.target_high, OperatingModes.ON)
        return mode


@attr.s
class Zone(Component):
    """This is representing a zone from the system.

    Note:
        If `rbr` (Room By Room) is True, the zone itself doesn't mean anything
        anymore, it means rooms are 'controlling' the zone.

    Args:
        active_function (ActiveFunction): Indicate what the zone is doing.
        rbr (bool): Room By Room, means the zone is controlled by ambisense.
        heating (ZoneHeating): Heating function of the zone.
        cooling (ZoneCooling): Cooling function of the zone.
        enabled: True if enabled
    """

    MODES = [
        OperatingModes.AUTO,
        OperatingModes.OFF,
        OperatingModes.DAY,
        OperatingModes.NIGHT,
        OperatingModes.QUICK_VETO,
        OperatingModes.ON,
        OperatingModes.TIME_CONTROLLED,
        OperatingModes.MANUAL,
    ]
    """List of mode that are applicable to zones component."""

    MIN_TARGET_HEATING_TEMP = constants.FROST_PROTECTION_TEMP
    """Min heating temperature that can be apply to a zone."""

    MIN_TARGET_COOLING_TEMP = 15
    """Min cooling temperature that can be apply to a zone."""

    MAX_TARGET_TEMP = constants.THERMOSTAT_MAX_TEMP
    """Max temperature that can be apply to a zone."""

    active_function = attr.ib(type=ActiveFunction, default=None)
    rbr = attr.ib(type=bool, default=False)
    heating = attr.ib(type=ZoneHeating, default=None)
    cooling = attr.ib(type=Optional[ZoneCooling], default=None)
    enabled = attr.ib(type=bool, default=True)

    @property
    def active_mode(self) -> Optional[ActiveMode]:
        if self.quick_veto:
            return ActiveMode(self.quick_veto.target, OperatingModes.QUICK_VETO)

        if self.active_function == ActiveFunction.COOLING and self.cooling:
            return self.cooling.active_mode
        if self.active_function == ActiveFunction.HEATING and self.heating:
            return self.heating.active_mode

        if self.heating:
            return self.heating.active_mode

        if self.cooling:
            return self.cooling.active_mode
        return None



"""Test for hot water."""
import unittest

from pymultimatic.model import OperatingModes, constants
from tests.conftest import _hotwater


class HotWaterTest(unittest.TestCase):
    """Test class."""

    def test_get_active_mode_on(self) -> None:
        """Test active mode on."""
        hot_water = _hotwater()
        hot_water.operating_mode = OperatingModes.ON

        active_mode = hot_water.active_mode

        self.assertEqual(OperatingModes.ON, active_mode.current)
        self.assertEqual(hot_water.target_high, active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_off(self) -> None:
        """Test active mode off."""
        hot_water = _hotwater()
        hot_water.operating_mode = OperatingModes.OFF

        active_mode = hot_water.active_mode

        self.assertEqual(OperatingModes.OFF, active_mode.current)
        self.assertEqual(constants.FROST_PROTECTION_TEMP, active_mode.target)
        self.assertIsNone(active_mode.sub)



"""Groups everything related to live report sensor."""
from datetime import date
from typing import Optional

import attr


@attr.s
class Report:
    """Represent a live report sensor."""

    id = attr.ib(type=str)
    value = attr.ib(type=float)
    name = attr.ib(type=str)
    unit = attr.ib(type=str)
    device_id = attr.ib(type=Optional[str], default=None)
    device_name = attr.ib(type=Optional[str], default=None)


@attr.s
class EmfReport:
    """Represent an emf report."""

    device_id = attr.ib(type=str)
    device_name = attr.ib(type=str)
    device_type = attr.ib(type=str)
    function = attr.ib(type=str)
    energyType = attr.ib(type=str)
    value = attr.ib(type=float)
    from_date = attr.ib(type=date)
    to_date = attr.ib(type=date)
